import sys
import os
import subprocess
import itertools
import uuid
import json
import datetime
import dateutil.parser
import types
import inspect
import operator
from enum import IntEnum
import sqlite3
from collections import defaultdict, OrderedDict

import networkx as nx
import dill


try:
	import simplevc
	_SUPPORT_SIMPLEVC = True
except:
	_SUPPORT_SIMPLEVC = False
	
###################################
# RMS Update events
###################################

class RMSUpdateEvent(IntEnum):
	MODIFY = 1
	INSERT = 2
	DELETE = 3
	CONTENTCHANGE = 4
	REPLACE = 5
	
###################################
# RMS Entries and related classes
###################################
class RMSEntryType(IntEnum):
	NONRMSENTRY = 0
	PIPE = 1
	RESOURCE = 2
	FILERESOURCE = 3
	TASK = 4
	UNRUNTASK = 5
	VIRTUALRESOURCE=6

class RMSEntry:
	def get_type(self):
		raise NotImplementedError()
	def get_id(self):
		raise NotImplementedError()
	def get_full_id(self):
		return (self.get_type(), self.get_id())
	def __key(self):
		return self.get_full_id()
	def __hash__(self):
		return hash(self.__key())
	def __eq__(self, other):
		if isinstance(other, RMSEntry):
			return self.__key() == other.__key()
		return NotImplemented

class Task(RMSEntry):
	def __init__(self, tid, pid, args, kwargs, return_values, output_files, begin_time, end_time, description, tags, info):
		self.tid = tid
		self.pid = pid
		self.args = args
		self.kwargs = kwargs
		self.return_values = return_values
		self.output_files = output_files
		self.begin_time = begin_time
		self.end_time = end_time
		self.description = description
		self.tags = tags
		self.info = info
	@property
	def input_pipes(self):
		return [arg for arg in self.args if isinstance(arg, Pipe)] + [arg for arg in self.kwargs.values() if isinstance(arg, Pipe)]
	@property
	def input_resources(self):
		return [arg for arg in self.args if isinstance(arg, Resource)] + [arg for arg in self.kwargs.values() if isinstance(arg, Resource)]
	@property
	def input_fileresources(self):
		return [arg for arg in self.args if isinstance(arg, FileResource)] + [arg for arg in self.kwargs.values() if isinstance(arg, FileResource)]
	@property
	def output_resources(self):
		return self.return_values
	@property
	def output_fileresources(self):
		return self.output_files
	@property
	def run_time(self):
		return self.end_time - self.begin_time
	def get_type(self):
		return RMSEntryType.TASK	
	def get_id(self):
		return self.tid
	
class UnrunTaskState(IntEnum):
	EDIT = 0
	LOCKED = 1
	PENDING = 2
	RUNNING = 3
	COMPLETE = 4
	ERROR = 5
class UnrunTask(RMSEntry):
	'''
	return_values could be None if not specified. 
	output_files could be None if not specified
	If the above two are specified (as a list of virtual resources), 
	an update will be applied to replace virtual resource to resource / fileresource
	'''
	def __init__(self, uid, pid, ba, return_values, output_files, 
				task_description="", task_tags=set(), task_info={},
				resource_description="", resource_tags=set(), resource_info={}, 
				fileresource_description="", fileresource_tags=set(), fileresource_info={}):
		self.uid = uid
		self.pid = pid
		self.ba = ba
		self.return_values = return_values
		self.output_files = output_files
		self.task_description = task_description
		self.task_tags = task_tags
		self.task_info = task_info
		self.resource_description = resource_description
		self.resource_tags = resource_tags
		self.resource_info = resource_info
		self.fileresource_description = fileresource_description
		self.fileresource_tags = fileresource_tags
		self.fileresource_info = fileresource_info
		self.replacement = None
		self.state = UnrunTaskState.EDIT
	@property
	def input_pipes(self):
		return [arg for arg in self.ba.args if isinstance(arg, Pipe)] + [arg for arg in self.ba.kwargs.values() if isinstance(arg, Pipe)]		
	@property
	def input_resources(self):
		return [arg for arg in self.ba.args if isinstance(arg, Resource)] + [arg for arg in self.ba.kwargs.values() if isinstance(arg, Resource)]
	@property
	def input_virtualresources(self):
		return [arg for arg in self.ba.args if isinstance(arg, VirtualResource)] + [arg for arg in self.ba.kwargs.values() if isinstance(arg, VirtualResource)]
	@property
	def input_fileresources(self):
		return [arg for arg in self.ba.args if isinstance(arg, FileResource)] + [arg for arg in self.ba.kwargs.values() if isinstance(arg, FileResource)]
	@property
	def output_resources(self):
		return [arg for arg in self.return_values if isinstance(arg, Resource)]
	@property
	def output_virtualresources(self):
		return [arg for arg in self.return_values if isinstance(arg, VirtualResource)]
	@property
	def output_fileresources(self):
		return self.output_files

	def get_type(self):
		return RMSEntryType.UNRUNTASK
	
	def get_id(self):
		return self.uid
	@property
	def ready(self):
		'''
		An unruntask is ready if all the arguments of ba is set properly.
		'''
		try:
			self.ba.signature.bind(*self.ba.args, **self.ba.kwargs)
		except TypeError:
			return False
		return True
class UnrunTaskParam:
	pass	
	# Remove associated parameter
	 
def _invalid_func():
	'''
	An indicator for invalid function when loading pipe from database.
	'''
	raise Exception("Invalid function")

class Pipe(RMSEntry):
	#def __init__(self, pid, func, return_tuple, return_volatile, output_func, is_deterministic, description, tags, info, rmsp):
	def __init__(self, pid, func_str, return_volatile, is_deterministic, module_name, func_name, output_func_str, description, tags, info, rms):
		self.pid = pid
		self._func_str = func_str
		self.return_volatile = return_volatile
		self.is_deterministic = is_deterministic
		self.module_name = module_name
		self.func_name = func_name		
		self._output_func_str = output_func_str
		self.description = description
		self.tags = tags
		self.info = info
		self.rms = rms
		self._func = None
		self._output_func = None
	
	@property
	def func(self):
		if self._func is None:
			try:
				loaded_obj = dill.loads(bytes.fromhex(self._func_str))
				if callable(loaded_obj):
					self._func = loaded_obj
				else:
					self._func = types.FunctionType(loaded_obj["__code__"], globals(), argdefs=loaded_obj["__defaults__"], closure=loaded_obj["__closure__"])
					# Note that the loaded obj will have rmsp.rmscore as module instead of __main__
					# Try to override that to __main__
					self._func.__module__ = "__main__"
					
			except:
				self._func = _invalid_func
		return self._func
	@property
	def output_func(self):
		if self._output_func_str is None:
			self._output_func = None
		elif self._output_func is None:
			try: 
				loaded_obj = dill.loads(bytes.fromhex(self._output_func_str))
				if isinstance(loaded_obj, types.FunctionType):
					self._output_func = loaded_obj
				else:
					self._output_func = types.FunctionType(loaded_obj["__code__"], globals(), argdefs=loaded_obj["__defaults__"], closure=loaded_obj["__closure__"])
				
			except:
				self._output_func = _invalid_func
		return self._output_func
			
			
	def __call__(self, *args, **kwargs):
		return self.rms.run(self, args, kwargs)
	def get_type(self):
		return RMSEntryType.PIPE	
	
	def get_id(self):
		return self.pid
	def __getstate__(self):
		d = dict(self.__dict__)
		d.pop("rms")
		return d
	def __setstate__(self, d):
		d = dict(d)
		d["rms"] = None
		self.__dict__ = d
	
class Resource(RMSEntry):
	def __init__(self, rid, tid, volatile, description, tags, info, content=None, has_content=True):
		self.rid = rid
		self.tid = tid
		self.volatile = volatile
		self.description = description
		self.tags = tags
		self.info = info
		self.has_content = has_content
		self._content = content
		
	@property
	def content(self):
		if self.has_content:
			if self.volatile:
				self.has_content = False
			return self._content
		else:
			raise Exception("No content is in resource")
		
	@content.setter
	def content(self, value):
		self.has_content = True
		self._content = value
		
	@content.deleter
	def content(self):
		self.has_content = False
		self._content = None
		
	def get_type(self):
		return RMSEntryType.RESOURCE	
	def get_id(self):
		return self.rid
	
	@property
	def content_type(self):
		if self.has_content:
			return type(self._content)
		else:
			raise Exception("No content is in resource")
	
class FileResource(RMSEntry):
	def __init__(self, fid, tid, file_path, md5, description, tags, info):
		self.fid = fid
		self.tid = tid
		self.file_path = file_path
		self.md5 = md5
		self.description = description
		self.tags = tags
		self.info = info
	def get_type(self):
		return RMSEntryType.FILERESOURCE	
		
	def get_id(self):
		return self.fid

class VirtualResource(RMSEntry):
	def __init__(self, vid, file_path=None):
		self.vid = vid
		self.uid = None
		self.replacement = None
		self.file_path = file_path
	def get_type(self):
		return RMSEntryType.VIRTUALRESOURCE	
		
	def get_id(self):
		return self.vid
	
# class VirtualFileResource(VirtualResource):
# 	def __init__(self, vid, file_path):
# 		super().__init__(vid)
# 		self.file_path = file_path
		
class ResourceNotReadyException(Exception):
	'''
	An exception that is called when accessing a resource without content.
	'''
	pass
	
class RMSTemplate():
	pass
class RMSLibrary():
	pass
class RMSTemplateJobStatus(IntEnum):
	PENDING = 0
	RUNNING = 1
	COMPLETE = 2
	ERROR = 3
class RMSTemplateJob():
	def __init__(self, tpid, func, args, kwargs):
		self.tpid = tpid
		self.func = func
		self.args = args
		self.kwargs = kwargs
		self.status = RMSTemplateJobStatus.PENDING
		self.unruntasks = []
		self.tasks = []
class VirtualModule():
	'''
	A virtual class to hold the imported module
	'''
	pass
	
##########################
# Useful SQL commands
##########################

# Inserts
def _sql_insert_fileresource(fileresource):
	cmds = []
	cmds.append(('''INSERT INTO files(fid, file_path, md5, description) VALUES(?,?,?,?)''',
						[fileresource.fid, fileresource.file_path, fileresource.md5, fileresource.description]))
	for t in fileresource.tags: 
		cmds.append(('''INSERT INTO file_tags(fid, tag_value) VALUES(?,?)''',
						[fileresource.fid, t]))
	for k, v in fileresource.info.items(): 
		cmds.append(('''INSERT INTO file_info(fid, info_key, info_value) VALUES(?,?,?)''',
						[fileresource.fid, k, v]))
	return cmds


def _sql_insert_task(task):
	cmds = []
	for i, arg in enumerate(task.args):
		if isinstance(arg, Resource):
			cmds.append(('''INSERT INTO tasks_args_resource(tid, arg_order, rid) VALUES(?,?,?)''',
							[task.tid, i, arg.rid]))
		elif isinstance(arg, FileResource):
			cmds.append(('''INSERT INTO tasks_args_file(tid, arg_order, fid) VALUES(?,?,?)''',
							[task.tid, i, arg.fid]))
		elif isinstance(arg, Pipe):
			cmds.append(('''INSERT INTO tasks_args_pipe(tid, arg_order, pid) VALUES(?,?,?)''',
							[task.tid, i, arg.pid]))
		else:
			cmds.append(('''INSERT INTO tasks_args_json(tid, arg_order, arg_value) VALUES(?,?,?)''',
							[task.tid, i, json.dumps(arg)]))
	for k, arg in task.kwargs.items():
		if isinstance(arg, Resource):
			cmds.append(('''INSERT INTO tasks_kwargs_resource(tid, arg_key, rid) VALUES(?,?,?)''',
							[task.tid, k, arg.rid]))
		elif isinstance(arg, FileResource):
			cmds.append(('''INSERT INTO tasks_kwargs_file(tid, arg_key, fid) VALUES(?,?,?)''',
							[task.tid, k, arg.fid]))
		elif isinstance(arg, Pipe):
			cmds.append(('''INSERT INTO tasks_kwargs_pipe(tid, arg_key, pid) VALUES(?,?,?)''',
							[task.tid, k, arg.pid]))
		else:
			cmds.append(('''INSERT INTO tasks_kwargs_json(tid, arg_key, arg_value) VALUES(?,?,?)''',
							[task.tid, k, json.dumps(arg)]))
	
	cmds.append(('''INSERT INTO tasks_returnvalue(tid, rid) VALUES(?,?)''',
							[task.tid, task.return_values[0].rid]))
	
	for i, f in enumerate(task.output_files): 
		cmds.append(('''INSERT INTO tasks_outputfiles(tid, forder, fid) VALUES(?,?,?)''',
								[task.tid, i, f.fid]))
		
		
	cmds.append(('''INSERT INTO tasks(tid, pid, begin_time, end_time, description) VALUES(?,?,?,?,?)''', 
		[task.tid, task.pid, task.begin_time.isoformat(), task.end_time.isoformat(), task.description]))
	for t in task.tags: 
		cmds.append(('''INSERT INTO task_tags(tid, tag_value) VALUES(?,?)''',
						[task.tid, t]))
	for k, v in task.info.items(): 
		cmds.append(('''INSERT INTO task_info(tid, info_key, info_value) VALUES(?,?,?)''',
						[task.tid, k, v]))
	return cmds
	
def _sql_insert_pipe(pipe):
	cmds = []
# 	cmds.append(('''INSERT INTO pipes(pid, func, return_volatile, is_deterministic, module_name, func_name, output_func, description) VALUES(?,?,?,?,?,?,?,?)''', 
# 		[pipe.pid, dill.dumps(pipe.func).hex(), int(pipe.return_volatile), int(pipe.is_deterministic), pipe.module_name, pipe.func_name, dill.dumps(pipe.output_func).hex() if pipe.output_func is not None else None, pipe.description]))
# 	cmds.append(('''INSERT INTO pipes(pid, func, return_volatile, is_deterministic, module_name, func_name, output_func, description) VALUES(?,?,?,?,?,?,?,?)''', 
# 		[pipe.pid, dill.dumps({"__code__":pipe.func.__code__, "__defaults__":pipe.func.__defaults__, "__closure__":pipe.func.__closure__}).hex(), int(pipe.return_volatile), int(pipe.is_deterministic), pipe.module_name, pipe.func_name, dill.dumps({"__code__":pipe.output_func.__code__, "__defaults__":pipe.output_func.__defaults__, "__closure__":pipe.output_func.__closure__}).hex() if pipe.output_func is not None else None, pipe.description]))
	#cmds.append(('''INSERT INTO pipes(pid, func, return_volatile, is_deterministic, module_name, func_name, output_func, description) VALUES(?,?,?,?,?,?,?,?)''', 
	#	[pipe.pid, _dumps_func(pipe.func), int(pipe.return_volatile), int(pipe.is_deterministic), pipe.module_name, pipe.func_name, _dumps_func(pipe.output_func) if pipe.output_func is not None else None, pipe.description]))
	cmds.append(('''INSERT INTO pipes(pid, func, return_volatile, is_deterministic, module_name, func_name, output_func, description) VALUES(?,?,?,?,?,?,?,?)''', 
		[pipe.pid, pipe._func_str, int(pipe.return_volatile), int(pipe.is_deterministic), pipe.module_name, pipe.func_name, pipe._output_func_str if pipe._output_func_str is not None else None, pipe.description]))
	for t in pipe.tags: 
		cmds.append(('''INSERT INTO pipe_tags(pid, tag_value) VALUES(?,?)''',
						[pipe.pid, t]))
	for k, v in pipe.info.items(): 
		cmds.append(('''INSERT INTO pipe_info(pid, info_key, info_value) VALUES(?,?,?)''',
						[pipe.pid, k, v]))
	return cmds
	
def _sql_insert_resource(resource):
	cmds = []
	cmds.append(('''INSERT INTO resources(rid, volatile, description) VALUES(?,?,?)''',
						[resource.rid, int(resource.volatile), resource.description]))
	for t in resource.tags: 
		cmds.append(('''INSERT INTO resource_tags(rid, tag_value) VALUES(?,?)''',
						[resource.rid, t]))
	for k, v in resource.info.items(): 
		cmds.append(('''INSERT INTO resource_info(rid, info_key, info_value) VALUES(?,?,?)''',
						[resource.rid, k, v]))
	return cmds

# Updates
def _sql_update_fileresource(fid, **kwargs):
	inserted_str = ",".join(map(lambda i:i[0] + "=?", kwargs.items()))
	return(f'''UPDATE files SET {inserted_str} WHERE fid = ?;''', [i[1] for i in kwargs.items()] + [fid])

def _sql_update_task(tid, **kwargs):
	inserted_str = ",".join(map(lambda i:i[0] + "=?", kwargs.items()))
	return(f'''UPDATE tasks SET {inserted_str} WHERE tid = ?;''', [i[1] for i in kwargs.items()] + [tid])

def _sql_update_pipe(pid, **kwargs):
	inserted_str = ",".join(map(lambda i:i[0] + "=?", kwargs.items()))
	return(f'''UPDATE pipes SET {inserted_str} WHERE pid = ?;''', [i[1] for i in kwargs.items()] + [pid])
def _sql_update_resource(rid, **kwargs):
	inserted_str = ",".join(map(lambda i:i[0] + "=?", kwargs.items()))
	return(f'''UPDATE resources SET {inserted_str} WHERE rid = ?;''', [i[1] for i in kwargs.items()] + [rid])
def _sql_update(full_id, **kwargs):
	if full_id[0] == RMSEntryType.FILERESOURCE:
		return _sql_update_fileresource(full_id[1], **kwargs)
	elif full_id[0] == RMSEntryType.TASK:
		return _sql_update_task(full_id[1], **kwargs)
	elif full_id[0] == RMSEntryType.PIPE:
		return _sql_update_pipe(full_id[1], **kwargs)
	elif full_id[0] == RMSEntryType.RESOURCE:
		return _sql_update_resource(full_id[1], **kwargs)
	else:
		raise Exception()
	
# Used in register, update, delete rmsobj and finish task
def _sql_execute_commands(sql, cmd_parameters):
	c = sql.cursor()
	c.execute("BEGIN")
	try:
		for cmd, parameters in cmd_parameters:
			c.execute(cmd, parameters)
		c.execute("COMMIT")
	except sql.Error as e:
		c.execute("ROLLBACK")
		raise Exception("Failure in SQL command execution")
	
	return c



###############################################
# Resources dump and fetch
###############################################
def _has_resource_content(path, resource):
	return os.path.exists(f"{path}/{resource.rid}")

def _load_resource_content(path, resource):
	'''
	Load the content of resource from the desired path, with the rid used as the file name
	'''
	if path is None:
		raise Exception("Resources folder is not available")
	if not os.path.exists(f"{path}/{resource.rid}"):
		raise Exception(f"Resource {resource.rid} is not available")
	with open(f"{path}/{resource.rid}", "rb") as file:
		resource.content = dill.load(file)
	
def _dump_resource_content(path, resource):
	'''
	Load the content of resource from the desired path, with the rid used as the file name
	'''
	if path is None:
		raise Exception("Resources folder is not available")
	if not resource.has_content:
		raise Exception("The resource has no content to save")
	with open(f"{path}/{resource.rid}", "wb") as file:
		dill.dump(resource.content, file)

#####################################
# Functions and signatures 
#####################################
def _inner_modify_func_code(c, remove_first_local):
	'''
	The functions try to modify func.__code__ so that two functions defined in different __main__ will remain the same.
	
	Try to remove all co_filename and co_firstlineno from the func codes.
	'''
	new_co_consts = []
	for i in range(len(c.co_consts)):
		if isinstance(c.co_consts[i], types.CodeType):
			new_co_consts.append(_inner_modify_func_code(c.co_consts[i], remove_first_local))
		elif remove_first_local and i > 0 and isinstance(c.co_consts[i - 1], types.CodeType):
			new_co_consts.append(c.co_consts[i][c.co_consts[i].find("<locals>") + 9:])
		else:
			new_co_consts.append(c.co_consts[i])
	new_co_consts = tuple(new_co_consts)
	if sys.version_info.major == 3 and sys.version_info.minor>=8:
		# 3.8 onwards
		return c.replace(co_consts=new_co_consts, co_filename="pipe", co_firstlineno = 1)
	else:
		# 3.7
		return types.CodeType(c.co_argcount,
		c.co_kwonlyargcount,
		c.co_nlocals,
		c.co_stacksize,
		c.co_flags,
		c.co_code,
		new_co_consts,
		c.co_names,
		c.co_varnames,
		"pipe",#c.co_filename,
		c.co_name,
		1,#c.co_firstlineno,
		c.co_lnotab,
		)
def _modify_func_code(c):
	c = _inner_modify_func_code(c, c.co_flags % 32 >= 16)
	new_co_flags = c.co_flags - c.co_flags % 32 + c.co_flags % 16
	if sys.version_info.major == 3 and sys.version_info.minor>=8:
		# 3.8 onwards
		return c.replace(co_flags=new_co_flags)
	else:
		# 3.7
		return types.CodeType(c.co_argcount,
		c.co_kwonlyargcount,
		c.co_nlocals,
		c.co_stacksize,
		new_co_flags, #c.co_flags,
		c.co_code,
		c.co_consts,
		c.co_names,
		c.co_varnames,
		c.co_filename,
		c.co_name,
		c.co_firstlineno,
		c.co_lnotab,
		)
	
def _get_func_signature(func):
	try:
		s = inspect.signature(func)
	except ValueError:
		# Try to use args and kwargs... 20210604 if no signature is found
		s = inspect.Signature([inspect.Parameter("args", inspect.Parameter.VAR_POSITIONAL), inspect.Parameter("kwargs", inspect.Parameter.VAR_KEYWORD)])
	return s

def _get_simplevc_func_signature(func):
	try:
		s = inspect.signature(func)
	except ValueError:
		s = inspect.Signature([inspect.Parameter("args", inspect.Parameter.VAR_POSITIONAL), inspect.Parameter("kwargs", inspect.Parameter.VAR_KEYWORD)])
	s = s.replace(parameters = tuple([p for p in s.parameters.values() if p.kind != inspect.Parameter.VAR_KEYWORD]) + (inspect.Parameter("version", inspect.Parameter.KEYWORD_ONLY),) + tuple([p for p in s.parameters.values() if p.kind == inspect.Parameter.VAR_KEYWORD]))
	return s

def _get_func_ba(func, args, kwargs, partial=False):
	# simplevc support
	if _SUPPORT_SIMPLEVC and func.__module__ in sys.modules and sys.modules[func.__module__] in simplevc.simplevc._registered_modules and func.__name__ in sys.modules[func.__module__]._method_dicts:
		module = sys.modules[func.__module__]
		if "version" not in kwargs or kwargs["version"] is None:
			version = module._version
		else:
			version = kwargs["version"]
		method_dict = module._method_dicts[func.__name__]
		version = simplevc.simplevc._get_last_version(sorted(method_dict.keys()), version)
		if version is None:
			raise Exception("Cannot find appropriate version")
		func =  method_dict[version]
		kwargs = {**kwargs, "version":version}
		if partial:
			ba = _get_simplevc_func_signature(func).bind_partial(*args, **kwargs)
		else:
			ba = _get_simplevc_func_signature(func).bind(*args, **kwargs)
	else:
		if partial:
			ba = _get_func_signature(func).bind_partial(*args, **kwargs)
		else:
			ba = _get_func_signature(func).bind(*args, **kwargs)
	ba.apply_defaults()
	return ba
def _dumps_func(func):
	if func.__module__ == "__main__":
		return dill.dumps({"__code__":func.__code__, "__defaults__":func.__defaults__, "__closure__":func.__closure__}).hex()
	else:
		return dill.dumps(func).hex()
#####################################
# Files MD5
#####################################
def _get_file_md5(f):
	if not os.path.exists(f):
		return None
	if os.path.isdir(f):
		return None
	try:
		return subprocess.check_output(["md5sum", f]).decode().split()[0]
	except:
		return None
	
#####################################
# Template Simulator
#####################################	
class TemplateSimulationResult:
	def __init__(self, rfdb, vfdb, unruntasks_db, virtualresources_db, pipes_db):
		self.rfdb = rfdb
		self.vfdb = vfdb
		self.unruntasks_db = unruntasks_db
		self.virtualresources_db = virtualresources_db
		self.pipes_db = pipes_db
		
class TemplateSimulator:
	def __init__(self):
		self.rfdb = {}
		self.vfdb = {}
		self.unruntasks_db = {}
		self.virtualresources_db = {}
		self.pipes_db = {}
		
	def _get_new_id(self):
		return uuid.uuid4().hex
		
	def create_virtualfileresource(self, file_path):
		vid = self._get_new_id()
		vr = VirtualResource(vid, file_path=file_path)
		self.vfdb[file_path] = vr
		self.virtualresources_db[vid] = vr 
		return vr
	def create_virtualresource(self,):
		vid = self._get_new_id()
		vr = VirtualResource(vid)
		self.virtualresources_db[vid] = vr
		return vr
	
	def create_unruntask(self, pid, args=[], kwargs={}, return_values=[], output_files=[],
						task_description="", task_tags=set(), task_info={},
						resource_description="", resource_tags=set(), resource_info={}, 
						fileresource_description="", fileresource_tags=set(), fileresource_info={}):
		uid = self._get_new_id()
		pipe = self.get_pipe(pid)
		ba = _get_func_ba(pipe.func, args, kwargs, partial=True)
		ut = UnrunTask(uid, pid, ba, return_values, output_files,
					task_description, task_tags, task_info,
					resource_description, resource_tags, resource_info,
					fileresource_description, fileresource_tags, fileresource_info
					)
		
		for vr in return_values:
			vr.uid = uid
		if output_files is not None:
			for vr in output_files:
				vr.uid = uid
		self.unruntasks_db[uid] = ut
		return ut
		
	
	def register_file(self, file_path):
		file_path = os.path.abspath(file_path)
		if file_path in self.rfdb:
			return self.rfdb[file_path]
		else:
			vr = self.create_virtualfileresource(file_path)
		self.rfdb[file_path] = vr
		return vr
	
	def file_from_path(self, file_path):
		file_path = os.path.abspath(file_path)
		if file_path in self.vfdb:
			return self.vfdb[file_path]
		else:
			vr = self.create_virtualfileresource(file_path)
		self.vfdb[file_path] = vr
		return vr
	
	def register_pipe(self, func, return_volatile=False, is_deterministic=True, output_func=None, description="", tags={}, pid=None, force=False):
		if not callable(func):
			raise Exception("func is not callable, and cannot be registered")
		info = {}
		if func.__module__ == "__main__":
			try:
				info["sourcecode"] = inspect.getsource(func)
			except:
				pass
			func.__code__ = _modify_func_code(func.__code__)
		if output_func is not None:
			if output_func.__module__ == "__main__":
				try:
					info["outputfunc_sourcecode"] = inspect.getsource(output_func)
				except:
					pass
				output_func.__code__ = _modify_func_code(output_func.__code__)
		e = dill.dumps(func)
		func = dill.loads(e)
		func_name = func.__name__
		module_name = func.__module__
		
		pid = self._get_new_id()
		pipe = Pipe(pid, dill.dumps(func).hex(), return_volatile, is_deterministic, module_name, func_name, None if output_func is None else dill.dumps(output_func).hex(), description, tags, info, self)
		self.pipes_db[pid] = pipe
		return pipe
	
	def get_pipe(self, pid):
		return self.pipes_db[pid]
	
	def run(self, pipe, args=(), kwargs={},
		   task_description="", task_tags=set(), task_info={},
				resource_description="", resource_tags=set(), resource_info={}, 
				fileresource_description="", fileresource_tags=set(), fileresource_info={}):
		uid = self._get_new_id()
		vr = self.create_virtualresource()
		if pipe.output_func is not None:
			try:
				def convert(arg):
					if isinstance(arg, VirtualResource) or isinstance(arg, FileResource):
						if arg.file_path is not None:
							return arg.file_path
						else:
							return arg
					elif isinstance(arg, Pipe):
						return arg.func
					else:
						return arg
				tmp_args = [convert(arg) for arg in args]
				tmp_kwargs = {key:convert(arg) for key, arg in kwargs.items()}
				
				virtual_output_files = pipe.output_func(*tmp_args, **tmp_kwargs)
				virtual_vrs = []
				for file_path in virtual_output_files: 
					file_path = os.path.abspath(file_path)
					if file_path in self.vfdb:
						tvr = self.vfdb[file_path]
					else:
						tvr = self.create_virtualfileresource(file_path)
						self.vfdb[file_path] = tvr
					virtual_vrs.append(tvr)
			except:
				virtual_vrs = []
		
		u = self.create_unruntask(pipe.pid, args, kwargs, [vr], virtual_vrs,
									task_description, task_tags, task_info,
									resource_description, resource_tags, resource_info,
									fileresource_description, fileresource_tags, fileresource_info
									)
		self.unruntasks_db[u.uid] = u
		return vr

	def get_result(self):
		return TemplateSimulationResult(self.rfdb, self.vfdb, self.unruntasks_db, self.virtualresources_db, self.pipes_db)
#####################################
# Others
#####################################	
def _group_by_first_id(data):
	d = defaultdict(list)
	for i in data:
		d[i[0]].append(i)
	return d	




#####################################
# RMS
#####################################	
class ResourceManagementSystem():
	'''
	A resource management system
	'''
	@property
	def sql(self):
		return sqlite3.connect(self.dbfile) 
	
	def __init__(self, dbfile, resource_dump_dir=None):
		self.dbfile = dbfile
		self.resources_db = {}
		self.pipes_db = {}
		self.tasks_db = {}
		self.fileresources_db = {}
		self.unruntasks_db = {}
		self.virtualresources_db = {}
# 		self.virtualfileresources_db = {}
		self.resource_dump_dir = resource_dump_dir
		self.tokens = {}
		self.rms_update_listeners = []
		self.vm = VirtualModule()
		self.scriptID = None
		self.allow_overwrite = False
		self.force_rerun = False
	############
	# RMS Info #
	############
	@property
	def database_id(self):
		dbid, = self.sql.cursor().execute("SELECT infovalue FROM metainfo WHERE infokey = 'dbid'").fetchone()
		return dbid
	def database_version(self):
		result = self.sql.cursor().execute("SELECT infovalue FROM metainfo WHERE infokey = 'dbversion'").fetchone()
		if result is None:
			return "1.0.0"
		else:
			return result[0]
	def set_scriptID(self, scriptID):
		self.scriptID = scriptID
	####################
	# RMS Update Event #
	####################
		
	def registerRMSUpdateListener(self, listener):
		'''
		Allow an external listener to listen to the broadcast from RMS.
		'''
		self.rms_update_listeners.append(listener)
		
	def fireRMSUpdateEvent(self, events):
		'''
		Fire RMS update events to the registered listeners.  
		'''
		for listener in self.rms_update_listeners:
			listener.onRMSUpdate(events)
			
	######################
	# Internal functions #
	######################
	
	def _obtain_resource_content(self, resource, autofetch=True):
		if not resource.has_content:
			try:
				_load_resource_content(self.resource_dump_dir, resource)
			except:
				if autofetch:
					# self.auto_fetch_resource_content([resource])
					if not resource.has_content:
						if resource.tid is None:
							raise Exception("Fail. Unable to retrieve resource tid.")
						task = self.get_task(resource.tid)
						pipe = self.get_pipe(task.pid)
						if not pipe.is_deterministic:
							raise Exception("Fail. Unable to rerun pipe because it is a non-deterministic pipe.")
						if not len(task.output_files) == 0:
							raise Exception("Fail. Re-retrieving the content may overwrite certain files.")
						if any(not os.path.exists(f.file_path) for f in task.input_fileresources):
							raise Exception("Fail. Some files are missing")
						for sub_resource in task.input_resources:
							self._obtain_resource_content(sub_resource, autofetch)
						raw_return_value = self._run_pipe(pipe, task.args, task.kwargs)
						resource.content = raw_return_value
						self.fireRMSUpdateEvent([(RMSUpdateEvent.CONTENTCHANGE, resource.get_full_id())])
					
					
					
				else:
					raise Exception("Resource content not available for " + resource.rid)
		return resource.content
	
	def _obtain_fileresource_file_path(self, fileresource):
		return fileresource.file_path
	
	def _obtain_pipe_func(self, pipe):
		return pipe.func
	
	def _obtain_rms_run_content(self, arg):
		if isinstance(arg, RMSEntry):
			if "overwritten" in arg.info or "deprecated" in arg.info:
				raise Exception("Overwritten or deprecated RMS obj cannot be used.")
			if isinstance(arg, Resource):
				return self._obtain_resource_content(arg)
			elif isinstance(arg, FileResource):
				return self._obtain_fileresource_file_path(arg)
			elif isinstance(arg, Pipe):
				return self._obtain_pipe_func(arg)
			else:
				raise Exception()
		else:
			return arg
		
	def _get_new_id(self):
		return uuid.uuid4().hex

	def _as_rmsobj(self, arg):
		if isinstance(arg, RMSEntry):
			rmsobj = arg
		else:
			rmsobj = self.get(arg)
		return rmsobj
	def _as_rmsid(self, arg):
		if isinstance(arg, RMSEntry):
			rmsid = arg.get_full_id()
		else:
			rmsid = arg
		return rmsid
	
	#############################
	# Users register a RMSEntry #
	#############################
	def register_pipe(self, func, return_volatile=False, is_deterministic=True, output_func=None, description="", tags={}, pid=None, force=False):
		'''
		'''
		if not callable(func):
			raise Exception("func is not callable, and cannot be registered")
		info = {}
		if func.__module__ == "__main__":
			try:
				info["sourcecode"] = inspect.getsource(func)
			except:
				pass
		if output_func is not None:
			if output_func.__module__ == "__main__":
				try:
					info["outputfunc_sourcecode"] = inspect.getsource(output_func)
				except:
					pass
				
		func = dill.loads(dill.dumps(func))
		if output_func is not None:
			output_func = dill.loads(dill.dumps(output_func))
		if func.__module__ == "__main__":
			if func.__code__.co_flags % 32 >= 16:
				if "<locals>" in func.__qualname__:
					func.__qualname__ = func.__qualname__[func.__qualname__.find("<locals>") + 9:]
			func.__code__ = _modify_func_code(func.__code__)
		if output_func is not None:
			if output_func.__module__ == "__main__":				
				if output_func.__code__.co_flags % 32 >= 16:
					if "<locals>" in output_func.__qualname__:
						output_func.__qualname__ = output_func.__qualname__[output_func.__qualname__.find("<locals>") + 9:]
				output_func.__code__ = _modify_func_code(output_func.__code__)
		func = dill.loads(dill.dumps(func))
		if output_func is not None:
			output_func = dill.loads(dill.dumps(output_func))
		func_name = func.__name__
		module_name = func.__module__
		cur_func_str = _dumps_func(func)#dill.dumps({"__code__":func.__code__, "__defaults__":func.__defaults__, "__closure__":func.__closure__}).hex()
		cur_output_func_str = _dumps_func(output_func) if output_func is not None else None #dill.dumps({"__code__":output_func.__code__, "__defaults__":output_func.__defaults__, "__closure__":output_func.__closure__}).hex() if output_func is not None else None
		c = self.sql.cursor()
		c.execute('''SELECT pid FROM pipes where return_volatile IS ? AND is_deterministic IS ? AND module_name IS ? AND func_name IS ?;''', 
				[int(return_volatile), int(is_deterministic), module_name, func_name])
		results = c.fetchall()
		if len(results) > 0:
			for old_pid, in results:
				old_pipe = self.get_pipe(old_pid)
				if (old_pipe._func_str == cur_func_str
					and ((old_pipe._output_func_str is None and cur_output_func_str is None)
						or (old_pipe.output_func is not None and cur_output_func_str is not None and old_pipe._output_func_str == cur_output_func_str))):
					return old_pipe
		if pid is None:
			pid = self._get_new_id()
		pipe = Pipe(pid, cur_func_str, return_volatile, is_deterministic, module_name, func_name, None if output_func is None else cur_output_func_str, description, tags, info, self)
		_sql_execute_commands(self.sql, _sql_insert_pipe(pipe))
		self.fireRMSUpdateEvent([(RMSUpdateEvent.INSERT, pipe.get_full_id())])
		return self.get_pipe(pid)
	
	
	def register_file(self, file_path, *, description="", tags=set(), tid=None, fid=None, force=False):
		'''
		'''
		file_path = os.path.abspath(file_path)
		try:
			existing_fileresource = self.file_from_path(file_path)
		except:
			existing_fileresource = None
		
		if force or (existing_fileresource is None):
			info = {}
			if fid is None:
				fid = self._get_new_id()
			if not os.path.exists(file_path):
				raise Exception(f"{file_path} does not exist")
			md5 = _get_file_md5(file_path)
			fileresource = FileResource(fid, tid, file_path, md5, description, tags, info)
			sql_cmds = []
			events = []
			sql_cmds.extend(_sql_insert_fileresource(fileresource))
			if existing_fileresource is not None:
				cur_time = datetime.datetime.now()
				sql_cmds.append(('INSERT INTO file_info(fid, info_key, info_value) VALUES(?,?,?)''', [existing_fileresource.fid, 'overwritten', cur_time.isoformat()]))
				events.append((RMSUpdateEvent.MODIFY, existing_fileresource.get_full_id()))
			_sql_execute_commands(self.sql, sql_cmds)
			self.fileresources_db[fid] = fileresource
			events.append((RMSUpdateEvent.INSERT, fileresource.get_full_id()))
			self.fireRMSUpdateEvent(events)
			return fileresource
		else:
			return existing_fileresource
	#############################
	# Users search for RMSEntry #
	#############################
	#
	# If users know the specific ID, see get
	#
	def file_from_path(self, file_path):
		file_path = os.path.abspath(file_path)
		c = self.sql.cursor()
		#c.execute('''SELECT fid FROM files where file_patfile_pathh = ? ORDER BY (SELECT end_time FROM tasks WHERE tid = tid) DESC;''', [file_path])
		c.execute('''SELECT fid FROM files where file_path = ? ORDER BY (SELECT end_time FROM tasks WHERE tid = tid) DESC;''', [file_path])
		raw_results = c.fetchall()
		results = [fid for fid, in raw_results if "overwritten" not in self.get_fileresource(fid).info and "deprecated" not in self.get_fileresource(fid).info]
		if len(results) > 0:
			if len(results) > 1:
				raise Exception("Found more than one file." + ",".join(results))
			fid = results[-1]
			return self.get_fileresource(fid)
		else:
			if len(raw_results) > 0:
				raise Exception("You have requested a deprecated / overwritten file") 
			else:
				raise Exception(f"File {file_path} is not registered.")
	
	def find_pipe(self, func):
		func_name = func.__name__
		module_name = func.__module__
		if not callable(func):
			raise Exception("func is not callable, and cannot be registered")
				
		func = dill.loads(dill.dumps(func))
			
		if func.__module__ == "__main__":
			if func.__code__.co_flags % 32 >= 16:
				if "<locals>" in func.__qualname__:
					func.__qualname__ = func.__qualname__[func.__qualname__.find("<locals>") + 9:]
			func.__code__ = _modify_func_code(func.__code__)

		c = self.sql.cursor()
		
		c.execute('''SELECT pid FROM pipes where module_name IS ? AND func_name IS ?;''', 
				[module_name, func_name])
		results = c.fetchall()
		filtered_results = []
		if len(results) > 0:
			for old_pid, in results:
				old_pipe = self.get_pipe(old_pid)
				if dill.dumps(old_pipe.func).hex() == dill.dumps(func).hex():
					filtered_results.append(old_pid)

		if len(filtered_results) > 0:
			if len(filtered_results) > 1:
				print("Warning: more than one pipe found.")
				print(filtered_results)
				raise Exception("Warning: more than one pipe found.")
			pid = filtered_results[0]
			return self.get_pipe(pid)
		raise Exception("Function not found")

	def find_tasks_by_io(self, iotype="io", fids=[], rids=[], pids=[], target_tids=None):
		'''
		Find tasks that have input and output
		
		If target tids are not supplied, all tasks are returned, which could take a long time to retrieve. 
		'''
		if target_tids is not None:
			s = ','.join(['\'' + tid + '\'' for tid in target_tids])
			tid_inclusion_string = f" AND tid IN ({s})"
		else: 
			tid_inclusion_string = ""
		cmds = []
		for fid in fids:
			if "i" in iotype:
				cmds.append(('''SELECT tid FROM tasks_args_file WHERE fid=?''', [fid]))
				cmds.append(('''SELECT tid FROM tasks_kwargs_file WHERE fid=?''', [fid]))
			if "o" in iotype:
				cmds.append(('''SELECT tid FROM tasks_outputfiles WHERE fid=?''', [fid]))
		for rid in rids:
			if "i" in iotype:
				cmds.append(('''SELECT tid FROM tasks_args_resource WHERE rid=?''', [rid]))
				cmds.append(('''SELECT tid FROM tasks_kwargs_resource WHERE rid=?''', [rid]))
			if "o" in iotype:
				cmds.append(('''SELECT tid FROM tasks_returnvalue WHERE rid=?''', [rid]))
		for pid in pids:
			if "i" in iotype:
				cmds.append(('''SELECT tid FROM tasks_args_pipe WHERE pid=?''', [pid]))
				cmds.append(('''SELECT tid FROM tasks_kwargs_pipe WHERE pid=?''', [pid]))
		sql_cmd = " UNION ".join([cmd[0] + tid_inclusion_string for cmd in cmds])
		sql_params = list(itertools.chain.from_iterable([cmd[1] for cmd in cmds]))
		tasks = [self.get_task(tid) for tid, in self.sql.cursor().execute(sql_cmd, sql_params)]
		return tasks
		
	def find_tasks_by_pipe(self, pids=[]):
		tids = set()
		for pid in pids:
			for tid, in self.sql.cursor().execute('''SELECT tid FROM tasks where pid=?;''', [pid]).fetchall():
				tids.add(tid)
		return [self.get_task(tid) for tid in tids]
	def find_tasks_by_pipe_and_args(self, pid, args, kwargs):
		# Split all resources
		resources = []
		files = []
		pipes = []
		jsons = []
		kw_resources = []
		kw_files = []
		kw_pipes = []
		kw_jsons = []
		for arg_order, arg in enumerate(args):
			if isinstance(arg, Resource):
				resources.append([arg_order, arg.rid])
			elif isinstance(arg, FileResource):
				files.append([arg_order, arg.fid])
			elif isinstance(arg, Pipe):
				pipes.append([arg_order, arg.pid])
			else:
				jsons.append([arg_order, json.dumps(arg)])
		for k, arg in kwargs.items():
			if isinstance(arg, Resource):
				kw_resources.append([k, arg.rid])
			elif isinstance(arg, FileResource):
				kw_files.append([k, arg.fid])
			elif isinstance(arg, Pipe):
				kw_pipes.append([k, arg.pid])
			else:
				kw_jsons.append([k, json.dumps(arg)])
	
		where_statements = []
		innerjoin_statements = []
		innerjoin_args = []
	
		# Create SQL statements for each table
		for tablename, groups, keyname, fieldname in [
			["tasks_args_resource", resources, "arg_order", "rid"], 
			["tasks_args_file", files, "arg_order", "fid"], 
			["tasks_args_pipe", pipes, "arg_order", "pid"], 
			["tasks_args_json", jsons, "arg_order", "arg_value"],
			["tasks_kwargs_resource", kw_resources, "arg_key", "rid"], 
			["tasks_kwargs_file", kw_files, "arg_key", "fid"], 
			["tasks_kwargs_pipe", kw_pipes, "arg_key", "pid"], 
			["tasks_kwargs_json", kw_jsons, "arg_key", "arg_value"],
	
		]:
			if len(groups) > 0:
				clause = " OR ".join([f"{keyname}=? AND {fieldname}=?" for _ in range(len(groups))])
				innerjoin_statements.append(f" INNER JOIN (SELECT tid, COUNT(*) AS count FROM {tablename} GROUP BY tid) {tablename}_prop ON tori.tid = {tablename}_prop.tid INNER JOIN (SELECT tid, COUNT(*) AS count FROM {tablename} WHERE {clause} GROUP BY tid) {tablename}_matches ON tori.tid = {tablename}_matches.tid AND {tablename}_prop.count = {tablename}_matches.count")
				innerjoin_args.extend(list(itertools.chain.from_iterable(groups)))
			else:
				where_statements.append(f" AND tid NOT IN (SELECT tid FROM {tablename})")
	
		statement = "SELECT tori.tid FROM (SELECT tid FROM tasks WHERE pid=? " + "".join(where_statements) + ") AS tori" + "".join(innerjoin_statements)
		args = [pid] + innerjoin_args
		tasks = [self.get_task(tid) for tid, in self.sql.cursor().execute(statement, args)]
		return tasks	

	def find_connected_objs(self, args, distance=-1, rmsobj_inclusion_criteria=lambda rmsobj:True, rmsobj_continue_search_criteria=lambda rmsobj:True, target_rmsobjs=None):
		return self._find_objs_by_dfs(self._get_next_connected_objs, args, distance, rmsobj_inclusion_criteria, rmsobj_continue_search_criteria, target_rmsobjs)
	def find_upstream_objs(self, args, distance=-1, rmsobj_inclusion_criteria=lambda rmsobj:True, rmsobj_continue_search_criteria=lambda rmsobj:True, target_rmsobjs=None):
		return self._find_objs_by_dfs(self._get_next_upstream_objs, args, distance, rmsobj_inclusion_criteria, rmsobj_continue_search_criteria, target_rmsobjs)
	def find_downstream_objs(self, args, distance=-1, rmsobj_inclusion_criteria=lambda rmsobj:True, rmsobj_continue_search_criteria=lambda rmsobj:True, target_rmsobjs=None):
		return self._find_objs_by_dfs(self._get_next_downstream_objs, args, distance, rmsobj_inclusion_criteria, rmsobj_continue_search_criteria, target_rmsobjs)
	
	def _find_objs_by_dfs(self, find_next_obj_func, args, distance, rmsobj_inclusion_criteria, rmsobj_continue_search_criteria, target_rmsobjs = None):
		# rmsobj_pool is the one to store the nodes to visit
		# visited is the one to store the output objs 
		
		rmsobjs = list(map(self._as_rmsobj, args))
		if target_rmsobjs is not None:
			target_rmsobjs = list(map(self._as_rmsobj, target_rmsobjs))
		rmsobj_pool = list()
		visited = set()
		for rmsobj in rmsobjs:
			rmsobj_pool.append((rmsobj, distance))		
			#visited.add(rmsobj)
		
		while len(rmsobj_pool) > 0:
			rmsobj, d = rmsobj_pool.pop() # Since we pop at the end.... it should be DFS...
			if d == 0:
				continue
			elif d > 0:
				d -= 1
			for new_rmsobj in find_next_obj_func(rmsobj, target_rmsobjs):
				if new_rmsobj not in visited and rmsobj_inclusion_criteria(new_rmsobj):
					visited.add(new_rmsobj)
					if rmsobj_continue_search_criteria(new_rmsobj):
						rmsobj_pool.append((new_rmsobj, d))
		return visited
		
	def _get_next_upstream_objs(self, rmsobj, target_rmsobjs=None):
		if rmsobj.get_type() == RMSEntryType.FILERESOURCE:
			if rmsobj.tid is None:
				return []
			objs = [self.get_task(rmsobj.tid)]
		elif rmsobj.get_type() == RMSEntryType.RESOURCE:
			if rmsobj.tid is None:
				return []
			objs = [self.get_task(rmsobj.tid)]
		elif rmsobj.get_type() == RMSEntryType.TASK:
			objs = rmsobj.input_resources + rmsobj.input_fileresources + rmsobj.input_pipes
		elif rmsobj.get_type() == RMSEntryType.UNRUNTASK:
			objs = rmsobj.input_resources + rmsobj.input_fileresources + rmsobj.input_pipes + rmsobj.input_virtualresources
		elif rmsobj.get_type() == RMSEntryType.VIRTUALRESOURCE:
			objs = [unruntask for unruntask in self.unruntasks_db.values() if rmsobj in unruntask.output_virtualresources]
			
			#if rmsobj.file_path is not None:
				#potential_unruntasks = [unruntask for unruntask in self.unruntasks_db.values() if rmsobj.file_path in self._guess_unruntask_output(unruntask)]
				#potential_unruntasks = [unruntask for unruntask in self.unruntasks_db.values() if rmsobj.file_path in unruntask.output_file]
				#objs += [r for r in potential_unruntasks if r not in objs] 
		elif rmsobj.get_type() == RMSEntryType.PIPE:
			objs = []
		else:
			raise Exception()
		if target_rmsobjs is not None:
			objs = [rmsobj for rmsobj in objs if rmsobj in target_rmsobjs]
		return objs
		
	def _get_next_downstream_objs(self, rmsobj, target_rmsobjs=None):
		target_tids = None if target_rmsobjs is None else [rmsobj.tid for rmsobj in target_rmsobjs if rmsobj.get_type() == RMSEntryType.TASK]
		if rmsobj.get_type() == RMSEntryType.FILERESOURCE:
			objs = self.find_tasks_by_io(iotype="i", fids=[rmsobj.fid], target_tids=target_tids) + [unruntask for unruntask in self.unruntasks_db.values() if rmsobj in unruntask.input_fileresources]
		elif rmsobj.get_type() == RMSEntryType.RESOURCE:
			objs = self.find_tasks_by_io(iotype="i", rids=[rmsobj.rid], target_tids=target_tids) + [unruntask for unruntask in self.unruntasks_db.values() if rmsobj in unruntask.input_resources]
			
		elif rmsobj.get_type() == RMSEntryType.TASK:
			objs = rmsobj.output_resources + rmsobj.output_fileresources
		elif rmsobj.get_type() == RMSEntryType.UNRUNTASK:
			objs = rmsobj.output_resources + rmsobj.output_fileresources + rmsobj.output_virtualresources
			file_paths = [obj.file_path for obj in objs if obj.file_path is not None]
# 			file_paths += self._guess_unruntask_output(rmsobj)
			file_paths = set(file_paths)
			potential_vrs = [virtualresource for virtualresource in self.virtualresources_db.values() if virtualresource.file_path is not None and virtualresource.file_path in file_paths]
			objs += [v for v in potential_vrs if v not in objs]
		elif rmsobj.get_type() == RMSEntryType.VIRTUALRESOURCE:
			objs = [unruntask for unruntask in self.unruntasks_db.values() if rmsobj in unruntask.input_virtualresources]
		elif rmsobj.get_type() == RMSEntryType.PIPE:
			objs = self.find_tasks_by_io(iotype="i", pids=[rmsobj.pid], target_tids=target_tids) + [unruntask for unruntask in self.unruntasks_db.values() if rmsobj in unruntask.input_pipes]
		else:
			raise Exception()
		#for i in self.unruntasks_db.values()
		if target_rmsobjs is not None:
			objs = [rmsobj for rmsobj in objs if rmsobj in target_rmsobjs]
			
		return objs
	def _get_next_connected_objs(self, rmsobj, target_rmsobjs=None):
		return set(self._get_next_upstream_objs(rmsobj, target_rmsobjs) + self._get_next_downstream_objs(rmsobj, target_rmsobjs))
	def find_upstream_rmsids(self, args, distance=-1, rmsobj_inclusion_criteria=lambda rmsobj:True, rmsobj_continue_search_criteria=lambda rmsobj:True, target_rmsids=None):
		return self.find_rmsids_by_bfs(self._get_multi_next_upstream_rmsids, args, distance, rmsobj_inclusion_criteria, rmsobj_continue_search_criteria, target_rmsids)
	def find_downstream_rmsids(self, args, distance=-1, rmsobj_inclusion_criteria=lambda rmsobj:True, rmsobj_continue_search_criteria=lambda rmsobj:True, target_rmsids=None):
		return self.find_rmsids_by_bfs(self._get_multi_next_downstream_rmsids, args, distance, rmsobj_inclusion_criteria, rmsobj_continue_search_criteria, target_rmsids)
					
	def find_rmsids_by_bfs(self, find_next_obj_func, args, distance, rmsid_inclusion_criteria, rmsid_continue_search_criteria, target_rmsids = None):
		'''
		rmsobj_inclusion_criteria:
		rmsobj_continue_search_criteria:
		'''
		# rmsobj_pool is the one to store the nodes to visit
		# visited is the one to store the output objs 
		
		rmsids = list(map(self._as_rmsid, args))
		if target_rmsids is not None:
			target_rmsids = list(map(self._as_rmsid, target_rmsids))
		visited = set()
		d = distance
		while len(rmsids) > 0 and d != 0:
			d -= 1
			new_rmsids = []
			for new_rmsid in find_next_obj_func(rmsids, target_rmsids):
				if new_rmsid not in visited and rmsid_inclusion_criteria(new_rmsid):
					visited.add(new_rmsid)
					if rmsid_continue_search_criteria(new_rmsid):
						new_rmsids.append(new_rmsid)
			rmsids = new_rmsids
		return visited
	def _get_multi_next_upstream_rmsids(self, rmsids, target_rmsids=None):
		c = self.sql.cursor()
		new_rmsids = []
		split_rmsid_list = {k:[] for k in [RMSEntryType.PIPE, RMSEntryType.FILERESOURCE, RMSEntryType.RESOURCE, RMSEntryType.TASK, RMSEntryType.UNRUNTASK, RMSEntryType.VIRTUALRESOURCE]}
		for rmsid in rmsids:
			split_rmsid_list[rmsid[0]].append(rmsid[1])
		rids = split_rmsid_list[RMSEntryType.RESOURCE]
		c.execute('''SELECT tid FROM tasks_returnvalue where rid IN ({});'''.format(",".join(["?"]*len(rids))), rids)
		new_rmsids.extend([(RMSEntryType.TASK, tid) for tid, in c.fetchall()])
		fids = split_rmsid_list[RMSEntryType.FILERESOURCE]
		c.execute('''SELECT tid FROM tasks_outputfiles where fid IN ({});'''.format(",".join(["?"]*len(fids))), fids)
		new_rmsids.extend([(RMSEntryType.TASK, tid) for tid, in c.fetchall()])
		
		tids = split_rmsid_list[RMSEntryType.TASK]
		c.execute('''SELECT rid FROM tasks_args_resource where tid IN ({});'''.format(",".join(["?"]*len(tids))), tids)
		new_rmsids.extend([(RMSEntryType.RESOURCE, rid) for rid, in c.fetchall()])
		c.execute('''SELECT fid FROM tasks_args_file where tid IN ({});'''.format(",".join(["?"]*len(tids))), tids)
		new_rmsids.extend([(RMSEntryType.FILERESOURCE, fid) for fid, in c.fetchall()])
		c.execute('''SELECT pid FROM tasks_args_pipe where tid IN ({});'''.format(",".join(["?"]*len(tids))), tids)
		new_rmsids.extend([(RMSEntryType.PIPE, pid) for pid, in c.fetchall()])
		c.execute('''SELECT rid FROM tasks_kwargs_resource where tid IN ({});'''.format(",".join(["?"]*len(tids))), tids)
		new_rmsids.extend([(RMSEntryType.RESOURCE, rid) for rid, in c.fetchall()])
		c.execute('''SELECT fid FROM tasks_kwargs_file where tid IN ({});'''.format(",".join(["?"]*len(tids))), tids)
		new_rmsids.extend([(RMSEntryType.FILERESOURCE, fid) for fid, in c.fetchall()])
		c.execute('''SELECT pid FROM tasks_kwargs_pipe where tid IN ({});'''.format(",".join(["?"]*len(tids))), tids)
		new_rmsids.extend([(RMSEntryType.PIPE, pid) for pid, in c.fetchall()])

		for uid in split_rmsid_list[RMSEntryType.UNRUNTASK]:
			rmsobj = self.unruntasks_db[uid]
			new_rmsids.extend([i.get_full_id() for i in rmsobj.input_resources + rmsobj.input_fileresources + rmsobj.input_pipes + rmsobj.input_virtualresources])
		for vid in split_rmsid_list[RMSEntryType.VIRTUALRESOURCE]:
			rmsobj = self.virtualresources_db[uid]
			new_rmsids.extend([unruntask.get_full_id() for unruntask in self.unruntasks_db.values() if rmsobj in unruntask.output_virtualresources])
			
		if target_rmsids is not None:
			new_rmsids = [rmsid for rmsid in new_rmsids if rmsid in target_rmsids]
		return new_rmsids

	def _get_multi_next_downstream_rmsids(self, rmsids, target_rmsids=None):
		def _as_type(t, l):
			return [(t, i) for i in l]
		split_rmsid_list = {k:[] for k in [RMSEntryType.PIPE, RMSEntryType.FILERESOURCE, RMSEntryType.RESOURCE, RMSEntryType.TASK, RMSEntryType.UNRUNTASK, RMSEntryType.VIRTUALRESOURCE]}
		for rmsid in rmsids:
			split_rmsid_list[rmsid[0]].append(rmsid[1])
		
		target_tids = None if target_rmsids is None else [rmsid[1] for rmsid in target_rmsids if rmsid[0] == RMSEntryType.TASK]
		
		objs = []
		# PIPE
		objs.extend(_as_type(RMSEntryType.TASK, self.find_tids_by_io(iotype="i", pids=split_rmsid_list[RMSEntryType.PIPE], target_tids=target_tids)))
		# RESOURCE
		objs.extend(_as_type(RMSEntryType.TASK, self.find_tids_by_io(iotype="i", rids=split_rmsid_list[RMSEntryType.RESOURCE], target_tids=target_tids)))
		# FILERESOURCE
		objs.extend(_as_type(RMSEntryType.TASK, self.find_tids_by_io(iotype="i", fids=split_rmsid_list[RMSEntryType.FILERESOURCE], target_tids=target_tids))) 
		# unruntask IO
		objs.extend([unruntask.get_full_id() for unruntask in self.unruntasks_db.values() if (any(f.fid in split_rmsid_list[RMSEntryType.FILERESOURCE] for f in unruntask.input_fileresources) or any(r.rid in split_rmsid_list[RMSEntryType.RESOURCE] for r in unruntask.input_resources) or any(p.pid in split_rmsid_list[RMSEntryType.PIPE] for p in unruntask.input_pipes))])
		# TASK		
		c = self.sql.cursor()
		tids = split_rmsid_list[RMSEntryType.TASK]
		c.execute('''SELECT rid FROM tasks_returnvalue where tid IN ({});'''.format(",".join(["?"]*len(tids))), tids)
		objs.extend([(RMSEntryType.RESOURCE, rid) for rid, in c.fetchall()])
		c.execute('''SELECT fid FROM tasks_outputfiles where tid IN ({}) ORDER BY forder ASC;'''.format(",".join(["?"]*len(tids))), tids) ###DDFDFSF
		objs.extend([(RMSEntryType.FILERESOURCE, fid) for fid, in c.fetchall()])		
		for uid in split_rmsid_list[RMSEntryType.UNRUNTASK]:
			rmsobj = self.unruntasks_db[uid]
			tmpobjs = rmsobj.output_resources + rmsobj.output_fileresources + rmsobj.output_virtualresources
			file_paths = [obj.file_path for obj in tmpobjs if obj.file_path is not None]
# 			file_paths += self._guess_unruntask_output(rmsobj)
			file_paths = set(file_paths)
			potential_vrs = [virtualresource for virtualresource in self.virtualresources_db.values() if virtualresource.file_path is not None and virtualresource.file_path in file_paths]
			tmpobjs += [v for v in potential_vrs if v not in tmpobjs]
			objs.extend([self._as_rmsid(obj) for obj in tmpobjs])
		for vid in split_rmsid_list[RMSEntryType.VIRTUALRESOURCE]:
			rmsobj = self.virtualresources_db[vid]
			objs.extend([unruntask.get_full_id() for unruntask in self.unruntasks_db.values() if rmsobj in unruntask.input_virtualresources])
		if target_rmsids is not None:
			objs = [rmsid for rmsid in objs if rmsid in target_rmsids]

		return objs		
		
	def find_tids_by_io(self, iotype="io", fids=[], rids=[], pids=[], target_tids=None):
		'''
		Find tasks that have input and output
		
		If target tids are not supplied, all tasks are returned, which could take a long time to retrieve. 
		'''
		if target_tids is not None:
			s = ','.join(['\'' + tid + '\'' for tid in target_tids])
			tid_inclusion_string = f" AND tid IN ({s})"
		else: 
			tid_inclusion_string = ""
		cmds = []

		if len(fids) > 0:
			if "i" in iotype:
				cmds.append(('''SELECT tid FROM tasks_args_file WHERE fid IN ({})'''.format(",".join(["?"]*len(fids))), fids))
				cmds.append(('''SELECT tid FROM tasks_kwargs_file WHERE fid IN ({})'''.format(",".join(["?"]*len(fids))), fids))
			if "o" in iotype:
				cmds.append(('''SELECT tid FROM tasks_outputfiles WHERE fid IN ({})'''.format(",".join(["?"]*len(fids))), fids))
		if len(rids) > 0:
			if "i" in iotype:
				cmds.append(('''SELECT tid FROM tasks_args_resource WHERE rid IN ({})'''.format(",".join(["?"]*len(rids))), rids))
				cmds.append(('''SELECT tid FROM tasks_kwargs_resource WHERE rid IN ({})'''.format(",".join(["?"]*len(rids))), rids))
			if "o" in iotype:
				cmds.append(('''SELECT tid FROM tasks_returnvalue WHERE rid IN ({})'''.format(",".join(["?"]*len(rids))), rids))
		if len(pids) > 0:
			if "i" in iotype:
				cmds.append(('''SELECT tid FROM tasks_args_pipe WHERE pid IN ({})'''.format(",".join(["?"]*len(pids))), pids))
				cmds.append(('''SELECT tid FROM tasks_kwargs_pipe WHERE pid IN ({})'''.format(",".join(["?"]*len(pids))), pids))
		sql_cmd = " UNION ".join([cmd[0] + tid_inclusion_string for cmd in cmds])
		sql_params = list(itertools.chain.from_iterable([cmd[1] for cmd in cmds]))
		return [tid for tid, in self.sql.cursor().execute(sql_cmd, sql_params)]
		
								
	############################
	# Users get RMSEntry by ID #
	############################
	#
	# Users get the RMSEntry by ID.
	# If the RMSEntry is already loaded in the memory (DB), use that copy.   
	# Otherwise, check from DB to load it to the memory
	#
	def _prefetch_entries(self, rmsids=[], fids=[], rids=[], pids=[], tids=[], refetch=False):
		'''
		Prefetch RMS Entries all at once instead of querying each node one by one
		'''
		c = self.sql.cursor()
		g = _group_by_first_id(rmsids)
		fids = fids + [i[1] for i in g[RMSEntryType.FILERESOURCE]]
		rids = rids + [i[1] for i in g[RMSEntryType.RESOURCE]]
		pids = pids + [i[1] for i in g[RMSEntryType.PIPE]]
		tids = tids + [i[1] for i in g[RMSEntryType.TASK]]
			
		# Pre-fetch all tasks first, since we may need to fetch extra resources, fileresources and pipes
		# TASK PART1
		c.execute('''SELECT * FROM tasks where tid IN ({});'''.format(",".join(["?"]*len(tids))), tids)
		pretask_results = c.fetchall()
		c.execute('''SELECT tid, tag_value FROM task_tags where tid IN ({});'''.format(",".join(["?"]*len(tids))), tids)
		pretask_tags_results = c.fetchall()
		c.execute('''SELECT tid, info_key, info_value FROM task_info where tid IN ({});'''.format(",".join(["?"]*len(tids))), tids)
		pretask_info_results = c.fetchall()
		c.execute('''SELECT tid, arg_order, arg_value FROM tasks_args_json where tid IN ({});'''.format(",".join(["?"]*len(tids))), tids)
		pre_args_json = c.fetchall()
		c.execute('''SELECT tid, arg_order, rid FROM tasks_args_resource where tid IN ({});'''.format(",".join(["?"]*len(tids))), tids)
		pre_args_resource = c.fetchall()
		c.execute('''SELECT tid, arg_order, fid FROM tasks_args_file where tid IN ({});'''.format(",".join(["?"]*len(tids))), tids)
		pre_args_fileresource = c.fetchall()
		c.execute('''SELECT tid, arg_order, pid FROM tasks_args_pipe where tid IN ({});'''.format(",".join(["?"]*len(tids))), tids)
		pre_args_pipe = c.fetchall()
		c.execute('''SELECT tid, arg_key, arg_value FROM tasks_kwargs_json where tid IN ({});'''.format(",".join(["?"]*len(tids))), tids)
		pre_kwargs_json = c.fetchall()
		c.execute('''SELECT tid, arg_key, rid FROM tasks_kwargs_resource where tid IN ({});'''.format(",".join(["?"]*len(tids))), tids)
		pre_kwargs_resource = c.fetchall()
		c.execute('''SELECT tid, arg_key, fid FROM tasks_kwargs_file where tid IN ({});'''.format(",".join(["?"]*len(tids))), tids)
		pre_kwargs_fileresource = c.fetchall()
		c.execute('''SELECT tid, arg_key, pid FROM tasks_kwargs_pipe where tid IN ({});'''.format(",".join(["?"]*len(tids))), tids)
		pre_kwargs_pipe = c.fetchall()
		c.execute('''SELECT tid, rid FROM tasks_returnvalue where tid IN ({});'''.format(",".join(["?"]*len(tids))), tids)
		pre_return_values = c.fetchall()
		c.execute('''SELECT tid, fid FROM tasks_outputfiles where tid IN ({}) ORDER BY forder ASC;'''.format(",".join(["?"]*len(tids))), tids) ###DDFDFSF
		pre_output_files = c.fetchall()


		# Add extra rids, fids, pids from tasks
		args_rids = [rid for tid, arg_order, rid in pre_args_resource]
		args_fids = [fid for tid, arg_order, fid in pre_args_fileresource]
		args_pids = [pid for tid, arg_order, pid in pre_args_pipe]

		kwargs_rids = [rid for tid, arg_key, rid in pre_kwargs_resource]
		kwargs_fids = [fid for tid, arg_key, fid in pre_kwargs_fileresource]
		kwargs_pids = [pid for tid, arg_key, pid in pre_kwargs_pipe]

		output_rids = [rid for tid, rid in pre_return_values]
		output_fids = [fid for tid, fid in pre_output_files]

		rids = list(set(rids + args_rids + kwargs_rids + output_rids))
		fids = list(set(fids + args_fids + kwargs_fids + output_fids))
		pids = list(set(pids + args_pids + kwargs_pids))
		if not refetch:
			rids = [rid for rid in rids if rid not in self.resources_db]
			fids = [fid for fid in fids if fid not in self.fileresources_db]
			pids = [pid for pid in pids if pid not in self.pipes_db]

		# RESOURCE
		c.execute('''SELECT * FROM resources where rid IN ({});'''.format(",".join(["?"]*len(rids))), rids)
		pre_results = c.fetchall()
		c.execute('''SELECT rid, tag_value FROM resource_tags where rid IN ({});'''.format(",".join(["?"]*len(rids))), rids)
		pre_tags_results = c.fetchall()
		c.execute('''SELECT rid, info_key, info_value FROM resource_info where rid IN ({});'''.format(",".join(["?"]*len(rids))), rids)
		pre_info_results = c.fetchall()
		c.execute('''SELECT rid, tid FROM tasks_returnvalue where rid IN ({});'''.format(",".join(["?"]*len(rids))), rids)
		pre_task_results = c.fetchall()
		
		results_dict, tags_results_dict, info_results_dict, task_results_dict = [_group_by_first_id(i) for i in [pre_results, pre_tags_results, pre_info_results, pre_task_results]]
		for rid in rids:
			results, tags_results, info_results, task_results = [i[rid] for i in [results_dict, tags_results_dict, info_results_dict, task_results_dict]]
			if len(results) != 1:
				raise Exception(results)
			if len(task_results) > 1:
				raise Exception()
			r = results[0]
			rid, volatile, description = (
				r[0],
				bool(r[1]),
				r[2]
			)
			tags = set(tag_value for rid, tag_value in tags_results)
			info = {k:v for rid, k, v in info_results}
			tid = task_results[0][1]
			if rid in self.resources_db: # Update when refetch
				resource = self.resources_db[rid]
				resource.rid=rid
				resource.tid=tid
				resource.volatile=volatile
				resource.description=description
				resource.tags=tags
				resource.info=info
			else:
				resource = Resource(rid, tid, volatile, description, tags, info, None, False)
				self.resources_db[resource.rid] = resource

		# FILERESOURCE
		c.execute('''SELECT * FROM files where fid IN ({});'''.format(",".join(["?"]*len(fids))), fids)
		pre_results = c.fetchall()
		c.execute('''SELECT fid, tag_value FROM file_tags where fid IN ({});'''.format(",".join(["?"]*len(fids))), fids)
		pre_tags_results = c.fetchall()
		c.execute('''SELECT fid, info_key, info_value FROM file_info where fid IN ({});'''.format(",".join(["?"]*len(fids))), fids)
		pre_info_results = c.fetchall()
		c.execute('''SELECT fid, tid FROM tasks_outputfiles where fid IN ({});'''.format(",".join(["?"]*len(fids))), fids)
		pre_task_results = c.fetchall()

		results_dict, tags_results_dict, info_results_dict, task_results_dict = [_group_by_first_id(i) for i in [pre_results, pre_tags_results, pre_info_results, pre_task_results]]
		for fid in fids:
			results, tags_results, info_results, task_results = [i[fid] for i in [results_dict, tags_results_dict, info_results_dict, task_results_dict]]
			if len(results) != 1:
				raise Exception("Error in getting files: " + fid)
			if len(task_results) > 1:
				raise Exception()
			r = results[0]
			fid, file_path, md5, description = (
				r[0],
				r[1],
				r[2],
				r[3]
			)
			tags = set(tag_value for fid, tag_value, in tags_results)
			info = {k:v for fid, k, v in info_results}
			if len(task_results) == 1:
				tid = task_results[0][1]
			else:
				tid = None
			if fid in self.fileresources_db:
				fileresource = self.fileresources_db[fid]
				fileresource.fid=fid
				fileresource.tid=tid
				fileresource.file_path=file_path
				fileresource.md5=md5
				fileresource.description=description
				fileresource.tags=tags
				fileresource.info=info
			else:
				fileresource = FileResource(fid, tid, file_path, md5, description, tags, info)
				self.fileresources_db[fileresource.fid] = fileresource





		# PIPE
		c.execute('''SELECT * FROM pipes where pid IN ({});'''.format(",".join(["?"]*len(pids))), pids)
		pre_results = c.fetchall()
		c.execute('''SELECT pid, tag_value FROM pipe_tags where pid IN ({});'''.format(",".join(["?"]*len(pids))), pids)
		pre_tags_results = c.fetchall()
		c.execute('''SELECT pid, info_key, info_value FROM pipe_info where pid IN ({});'''.format(",".join(["?"]*len(pids))), pids)
		pre_info_results = c.fetchall()

		results_dict, tags_results_dict, info_results_dict = [_group_by_first_id(i) for i in [pre_results, pre_tags_results, pre_info_results]]
		for pid in pids:
			results, tags_results, info_results = [i[pid] for i in [results_dict, tags_results_dict, info_results_dict]]
			if len(results) != 1:
				raise Exception()

			r = results[0]
			pid, func_str, return_volatile, is_deterministic, module_name, func_name, output_func_str, description = (
				r[0],
				r[1],
				bool(r[2]),
				bool(r[3]),
				r[4],
				r[5],
				r[6],
				r[7]
			)
			tags = set(tag_value for pid, tag_value, in tags_results)
			info = {k:v for pid, k, v in info_results}
			if pid in self.pipes_db:
				pipe = self.pipes_db[pid]
				pipe.pid = pid
				pipe.func_str = func_str
				pipe.return_volatile = return_volatile
				pipe.is_deterministic = is_deterministic
				pipe.module_name = module_name
				pipe.func_name = func_name
				pipe.output_func_str = output_func_str
				pipe.description = description
				pipe.tags = tags
				pipe.info = info
				pipe._func = None
				pipe._output_func = None
			else:
				pipe = Pipe(pid, func_str, return_volatile, is_deterministic, module_name, func_name, output_func_str, description, tags, info, self)
				self.pipes_db[pipe.pid] = pipe


		# TASK PART2
		results_dict, tags_results_dict, info_results_dict, args_json_dict, args_resource_dict, args_fileresource_dict, args_pipe_dict, kwargs_json_dict, kwargs_resource_dict, kwargs_fileresource_dict, kwargs_pipe_dict, return_values_dict, output_files_dict = [_group_by_first_id(i) for i in [pretask_results, pretask_tags_results, pretask_info_results, pre_args_json, pre_args_resource, pre_args_fileresource, pre_args_pipe, pre_kwargs_json, pre_kwargs_resource, pre_kwargs_fileresource, pre_kwargs_pipe, pre_return_values, pre_output_files]]
		for tid in tids:
			results, tags_results, info_results, args_json, args_resource, args_fileresource, args_pipe, kwargs_json, kwargs_resource, kwargs_fileresource, kwargs_pipe, return_values, output_files = [i[tid] for i in [results_dict, tags_results_dict, info_results_dict, args_json_dict, args_resource_dict, args_fileresource_dict, args_pipe_dict, kwargs_json_dict, kwargs_resource_dict, kwargs_fileresource_dict, kwargs_pipe_dict, return_values_dict, output_files_dict]]
			args_json = [(arg_order, json.loads(arg_value)) for tid, arg_order, arg_value in args_json]
			args_resource = [(arg_order, self.get_resource(rid)) for tid, arg_order, rid in args_resource]
			args_fileresource = [(arg_order, self.get_fileresource(fid)) for tid, arg_order, fid in args_fileresource]
			args_pipe = [(arg_order, self.get_pipe(pid)) for tid, arg_order, pid in args_pipe]
			args = list(map(operator.itemgetter(1), sorted(itertools.chain.from_iterable([args_json, args_resource, args_fileresource, args_pipe]), key=operator.itemgetter(0))))
			kwargs_json = {arg_key:json.loads(arg_value) for tid, arg_key, arg_value in kwargs_json}
			kwargs_resource = {arg_key:self.get_resource(rid) for tid, arg_key, rid in kwargs_resource}
			kwargs_fileresource = {arg_key:self.get_fileresource(fid) for tid, arg_key, fid in kwargs_fileresource}
			kwargs_pipe = {arg_key:self.get_pipe(pid) for tid, arg_key, pid in kwargs_pipe}
			return_values = [self.get_resource(rid) for tid, rid, in return_values]
			output_files = [self.get_fileresource(fid) for tid, fid, in output_files]
			kwargs = {**kwargs_json, **kwargs_resource, **kwargs_fileresource, **kwargs_pipe}
			r = results[0]
			tid, pid, begin_time, end_time, description = (
				r[0], 
				r[1],
				dateutil.parser.isoparse(r[2]),
				dateutil.parser.isoparse(r[3]),
				r[4]
			)
			tags = set(tag_value for tid, tag_value, in tags_results)
			info = {k:v for tid, k, v in info_results}

			if tid in self.tasks_db:
				task = self.tasks_db[tid]
				task.tid=tid
				task.pid=pid
				task.args=args
				task.kwargs=kwargs
				task.return_values=return_values
				task.output_files=output_files
				task.begin_time=begin_time
				task.end_time=end_time
				task.description=description
				task.tags=tags
				task.info=info
			else:
				task = Task(tid, pid, args, kwargs, return_values, output_files, begin_time, end_time, description, tags, info)
				self.tasks_db[task.tid] = task
				
				
				
				
	def has(self, arg):
		try:
			self.get(arg)
			return True
		except:
			return False
	def get(self, arg, refetch=False):
		if isinstance(arg, RMSEntry):
			return arg
		rmsid = arg
		if rmsid[0] == RMSEntryType.PIPE:
			return self.get_pipe(rmsid[1], refetch)
		elif rmsid[0] == RMSEntryType.RESOURCE:
			return self.get_resource(rmsid[1], refetch)
		elif rmsid[0] == RMSEntryType.FILERESOURCE:
			return self.get_fileresource(rmsid[1], refetch)
		elif rmsid[0] == RMSEntryType.TASK:
			return self.get_task(rmsid[1], refetch)
		elif rmsid[0] == RMSEntryType.UNRUNTASK:
			return self.get_unruntask(rmsid[1])
		elif rmsid[0] == RMSEntryType.VIRTUALRESOURCE:
			return self.get_virtualresource(rmsid[1])
		else:
			raise Exception()
		
	def get_pipe(self, pid, refetch=False):
		if pid in self.pipes_db and not refetch:
			return self.pipes_db[pid]		
		c = self.sql.cursor()
		c.execute('''SELECT * FROM pipes where pid = ?;''', [pid])
		results = c.fetchall()
		c.execute('''SELECT tag_value FROM pipe_tags where pid = ?;''', [pid])
		tags_results = c.fetchall()
		c.execute('''SELECT info_key, info_value FROM pipe_info where pid = ?;''', [pid])
		info_results = c.fetchall()
		if len(results) != 1:
			raise Exception()
		
		r = results[0]
		pid, func_str, return_volatile, is_deterministic, module_name, func_name, output_func_str, description = (
			r[0],
			r[1],
			bool(r[2]),
			bool(r[3]),
			r[4],
			r[5],
			r[6],
			r[7]
		)
		tags = set(tag_value for tag_value, in tags_results)
		info = {k:v for k, v in info_results}
		if pid in self.pipes_db:
			raise Exception()
		else:
			pipe = Pipe(pid, func_str, return_volatile, is_deterministic, module_name, func_name, output_func_str, description, tags, info, self)
			self.pipes_db[pipe.pid] = pipe
		return pipe
	
	def get_resource(self, rid, refetch=False):
		if rid in self.resources_db and not refetch:
			return self.resources_db[rid]
		c = self.sql.cursor()
		c.execute('''SELECT * FROM resources where rid = ?;''', [rid])
		results = c.fetchall()
		c.execute('''SELECT tag_value FROM resource_tags where rid = ?;''', [rid])
		tags_results = c.fetchall()
		c.execute('''SELECT info_key, info_value FROM resource_info where rid = ?;''', [rid])
		info_results = c.fetchall()
		c.execute('''SELECT tid FROM tasks_returnvalue where rid = ?;''', [rid])
		task_results = c.fetchall()
		
		if len(results) != 1:
			raise Exception(results)
		if len(task_results) > 1:
			raise Exception()
		elif len(task_results) == 1:
			tid, = task_results[0]
		else:
			tid = None
		r = results[0]
		rid, volatile, description = (
			r[0],
			bool(r[1]),
			r[2]
		)
		tags = set(tag_value for tag_value, in tags_results)
		info = {k:v for k, v in info_results}
		
		if rid in self.resources_db:
			resource = self.resources_db[rid]
			resource.rid=rid
			resource.tid=tid
			resource.volatile=volatile
			resource.description=description
			resource.tags=tags
			resource.info=info
		else:
			resource = Resource(rid, tid, volatile, description, tags, info, None, False)
			self.resources_db[resource.rid] = resource
		return resource

	def get_fileresource(self, fid, refetch=False):
		if fid in self.fileresources_db and not refetch:
			return self.fileresources_db[fid]
		c = self.sql.cursor()
		c.execute('''SELECT * FROM files where fid = ?;''', [fid])
		results = c.fetchall()
		c.execute('''SELECT tag_value FROM file_tags where fid = ?;''', [fid])
		tags_results = c.fetchall()
		c.execute('''SELECT info_key, info_value FROM file_info where fid = ?;''', [fid])
		info_results = c.fetchall()
		c.execute('''SELECT tid FROM tasks_outputfiles where fid = ?;''', [fid])
		task_results = c.fetchall()
		if len(results) != 1:
			raise Exception("Error in getting files: " + fid)
		if len(task_results) > 1:
			raise Exception()
		elif len(task_results) == 1:
			tid, = task_results[0]
		else:
			tid = None
		r = results[0]
		fid, file_path, md5, description = (
			r[0],
			r[1],
			r[2],
			r[3]
		)
		tags = set(tag_value for tag_value, in tags_results)
		info = {k:v for k, v in info_results}
		
		if fid in self.fileresources_db:
			fileresource = self.fileresources_db[fid]
			fileresource.fid=fid
			fileresource.tid=tid
			fileresource.file_path=file_path
			fileresource.md5=md5
			fileresource.description=description
			fileresource.tags=tags
			fileresource.info=info
		else:
			fileresource = FileResource(fid, tid, file_path, md5, description, tags, info)
			self.fileresources_db[fileresource.fid] = fileresource
		return fileresource

	def get_task(self, tid, refetch=False):
		
		if tid in self.tasks_db and not refetch:
			return self.tasks_db[tid]
		c = self.sql.cursor()
		c.execute('''SELECT * FROM tasks where tid = ?;''', [tid])
		results = c.fetchall()
		if len(results) != 1:
			raise Exception()
		c.execute('''SELECT tag_value FROM task_tags where tid = ?;''', [tid])
		tags_results = c.fetchall()
		c.execute('''SELECT info_key, info_value FROM task_info where tid = ?;''', [tid])
		info_results = c.fetchall()
		
		c.execute('''SELECT arg_order, arg_value FROM tasks_args_json where tid = ?;''', [tid])
		args_json = [(arg_order, json.loads(arg_value)) for arg_order, arg_value in c.fetchall()]
		c.execute('''SELECT arg_order, rid FROM tasks_args_resource where tid = ?;''', [tid])
		args_resource = [(arg_order, self.get_resource(rid)) for arg_order, rid in c.fetchall()]
		c.execute('''SELECT arg_order, fid FROM tasks_args_file where tid = ?;''', [tid])
		args_fileresource = [(arg_order, self.get_fileresource(fid)) for arg_order, fid in c.fetchall()]
		c.execute('''SELECT arg_order, pid FROM tasks_args_pipe where tid = ?;''', [tid])
		args_pipe = [(arg_order, self.get_pipe(pid)) for arg_order, pid in c.fetchall()]
		args = list(map(operator.itemgetter(1), sorted(itertools.chain.from_iterable([args_json, args_resource, args_fileresource, args_pipe]), key=operator.itemgetter(0))))

		c.execute('''SELECT arg_key, arg_value FROM tasks_kwargs_json where tid = ?;''', [tid])
		kwargs_json = {arg_key:json.loads(arg_value) for arg_key, arg_value in c.fetchall()}
		c.execute('''SELECT arg_key, rid FROM tasks_kwargs_resource where tid = ?;''', [tid])
		kwargs_resource = {arg_key:self.get_resource(rid) for arg_key, rid in c.fetchall()}
		c.execute('''SELECT arg_key, fid FROM tasks_kwargs_file where tid = ?;''', [tid])
		kwargs_fileresource = {arg_key:self.get_fileresource(fid) for arg_key, fid in c.fetchall()}
		c.execute('''SELECT arg_key, pid FROM tasks_kwargs_pipe where tid = ?;''', [tid])
		kwargs_pipe = {arg_key:self.get_pipe(pid) for arg_key, pid in c.fetchall()}
		kwargs = {**kwargs_json, **kwargs_resource, **kwargs_fileresource, **kwargs_pipe}
		
		c.execute('''SELECT rid FROM tasks_returnvalue where tid = ?;''', [tid])
		return_values = [self.get_resource(rid) for rid, in c.fetchall()]
		
		c.execute('''SELECT fid FROM tasks_outputfiles where tid = ? ORDER BY forder ASC;''', [tid])
		output_files = [self.get_fileresource(fid) for fid, in c.fetchall()]
		
		r = results[0]
		tid, pid, begin_time, end_time, description = (
			r[0], 
			r[1],
			dateutil.parser.isoparse(r[2]),
			dateutil.parser.isoparse(r[3]),
			r[4]
		)
		tags = set(tag_value for tag_value, in tags_results)
		info = {k:v for k, v in info_results}
		
		if tid in self.tasks_db:
			task = self.tasks_db[tid]
			task.tid=tid
			task.pid=pid
			task.args=args
			task.kwargs=kwargs
			task.return_values=return_values
			task.output_files=output_files
			task.begin_time=begin_time
			task.end_time=end_time
			task.description=description
			task.tags=tags
			task.info=info
		else:
			task = Task(tid, pid, args, kwargs, return_values, output_files, begin_time, end_time, description, tags, info)
			self.tasks_db[task.tid] = task
		return task
	def get_unruntask(self, uid):
		if uid in self.unruntasks_db:
			return self.unruntasks_db[uid]
		raise Exception()
	def get_virtualresource(self, vid):
		if vid in self.virtualresources_db:
			return self.virtualresources_db[vid]
		raise Exception()

	# Methods for loading RMSEntry from DB to memory
	def parse_row_as_pipe_param(self, row):
		return (row[0], dill.loads(bytes.fromhex(row[1])), bool(row[2]), json.loads(row[3]), dill.loads(bytes.fromhex(row[4])) if row[4] is not None else None, bool(row[5]), row[6], row[7], json.loads(row[8])) 
	def parse_row_as_fileresource_param(self, row):
		return (row[0], row[1], row[2], row[3], row[4], json.loads(row[5]))
		fileresource = FileResource(row[0], row[1], row[2], row[3], row[4], json.loads(row[5]))
		self.fileresources_db[row[0]] = fileresource
		return fileresource 

	def parse_row_as_resource_param(self, row):
		return (row[0], row[1], bool(row[2]), row[3], row[4], json.loads(row[5]))
	def parse_row_as_task_param(self, row):
		convert_funcs = {"r": lambda s: self.get_resource(s),
						 "f": lambda s: self.get_fileresource(s),
						 "i": lambda s: self.get_pipe(s),
						 "p": json.loads}
		
		args = [convert_funcs[arg[0]](arg[1:]) for arg in json.loads(row[2])]
		kwargs = {k:convert_funcs[arg[0]](arg[1:]) for k, arg in json.loads(row[3]).items()}
		return_values = [self.get_resource(arg) for arg in json.loads(row[4])]
		output_files = [self.get_fileresource(arg) for arg in json.loads(row[5])]
		return (row[0], row[1], args, kwargs, return_values, output_files, dateutil.parser.isoparse(row[6]), dateutil.parser.isoparse(row[7]), row[8], row[9], json.loads(row[10]))

	def save_resource_content(self, resource):
		_dump_resource_content(self.resource_dump_dir, resource)
	
	def get_auto_fetch_resource_content_info(self, resources, allow_non_deterministic=False):
		resources = list(map(self._as_rmsobj, resources))
		dry_run_resources = set()
		dry_run_tasks = set()
		for resource in resources:
			if resource not in dry_run_resources and not resource.has_content:
				if resource.tid is None:
					raise Exception("Fail. Unable to retrieve resource tid.")
				task = self.get_task(resource.tid)
				pipe = self.get_pipe(task.pid)
				if not allow_non_deterministic and not pipe.is_deterministic:
					raise Exception("Fail. Unable to rerun pipe because it is a non-deterministic pipe.")
				if not len(task.output_files) == 0:
					raise Exception("Fail. Re-retrieving the content may overwrite certain files.")
				if any(not os.path.exists(f.file_path) for f in task.input_fileresources):
					raise Exception("Fail. Some files are missing")
				
				required_tasks, required_resources = self.get_auto_fetch_resource_content_info(task.input_resources, allow_non_deterministic=allow_non_deterministic)
				dry_run_tasks.update(required_tasks)
				dry_run_resources.update(required_resources)
				dry_run_tasks.add(task)
				dry_run_resources.add(resource)
		return dry_run_tasks, dry_run_resources
							
	def auto_fetch_resource_content(self, resources, allow_non_deterministic=False):
		'''
		Automatically fetch resource content if not dumped previously. Rerun all necessary tasks to obtain the resources
		'''
		resources = list(map(self._as_rmsobj, resources))
		for resource in resources:
			if not resource.has_content:
				if resource.tid is None:
					raise Exception("Fail. Unable to retrieve resource tid.")
				task = self.get_task(resource.tid)
				pipe = self.get_pipe(task.pid)
				if not allow_non_deterministic and not pipe.is_deterministic:
					raise Exception("Fail. Unable to rerun pipe because it is a non-deterministic pipe.")
				if not len(task.output_files) == 0:
					raise Exception("Fail. Re-retrieving the content may overwrite certain files.")
				self.auto_fetch_resource_content(task.input_resources, allow_non_deterministic=allow_non_deterministic)
				if any(not os.path.exists(f.file_path) for f in task.input_fileresources):
					raise Exception("Fail. Some files are missing")
				
				raw_return_value = self._run_pipe(pipe, task.args, task.kwargs)
				resource.content = raw_return_value
				self.fireRMSUpdateEvent([(RMSUpdateEvent.CONTENTCHANGE, resource.get_full_id())])
	####################
	# Users run a task #
	####################
	def compute_output_files(self, pipe, func_args, func_kwargs):
		output_file_paths = pipe.output_func(*func_args, **func_kwargs) if pipe.output_func is not None else []
		output_file_paths = [os.path.abspath(path) for path in output_file_paths] 
		return output_file_paths
	
	def exist_output_files(self, output_files):
		overlapped_fids = []
		for output_file in output_files:
			c = self.sql.cursor()
			c.execute('''SELECT fid FROM files where file_path = ?;''', [output_file])
			overlapped_fids.extend([fid for fid, in c.fetchall()])
		return len(overlapped_fids) > 0
	
	def _compile_files_md5_from_pipe_ba(self, pipe, ba):
		func_args, func_kwargs = self._ba_rms_to_func_args_kwargs(ba)
		# Check for output files
		output_file_paths = pipe.output_func(*func_args, **func_kwargs) if pipe.output_func is not None else []
		output_file_paths = [os.path.abspath(path) for path in output_file_paths] 
		problematic_paths = [p for p in output_file_paths if not os.path.exists(p)]
		if len(problematic_paths) > 0:
			print("Warning: the following output files are not detected:")
			print(problematic_paths)
			output_file_paths = [p for p in output_file_paths if os.path.exists(p)]
		return OrderedDict([[file_path,_get_file_md5(file_path)] for file_path in output_file_paths])
	def register_finished_task(self, raw_return_value, begin_time, end_time, ba, pipe, 
							task_description="", task_tags=[], task_info={},
							resource_description="", resource_tags=[], resource_info={},
							fileresource_description="", fileresource_tags=[], fileresource_info={},
							precompiled_files_md5=None
							):
		'''
		Register a finished task, assuming the task has completed successfully. 
		'''
# 			
		if precompiled_files_md5 is None:
			file_path_md5_dict = self._compile_files_md5_from_pipe_ba(pipe, ba)
		else:
			file_path_md5_dict = precompiled_files_md5
		output_file_paths = list(file_path_md5_dict.keys())
		return_value_rid = self._get_new_id()
		output_file_fids = [self._get_new_id() for _ in output_file_paths]
		tid = self._get_new_id()
		
		# Create resource, fileresource and task
		new_resources = [Resource(return_value_rid, tid, pipe.return_volatile, resource_description, resource_tags, resource_info, raw_return_value)]
		
# 		new_fileresources = [FileResource(fid, tid, file_path, _get_file_md5(file_path), fileresource_description, fileresource_tags, fileresource_info) for fid, file_path in zip(output_file_fids, output_file_paths)]
		new_fileresources = [FileResource(fid, tid, file_path, file_path_md5_dict[file_path], fileresource_description, fileresource_tags, fileresource_info) for fid, file_path in zip(output_file_fids, output_file_paths)]
		task = Task(tid, pipe.pid, ba.args, ba.kwargs, new_resources, new_fileresources, begin_time, end_time, task_description, task_tags, task_info)
		
		# Find files that are overwritten
		# Even if deprecated, it should be marked overwritten
		old_fids = []
		for fileresource in new_fileresources:
			c = self.sql.cursor()
			c.execute('''SELECT fid FROM files where file_path = ?;''', [fileresource.file_path])
			old_fids.extend([fid for fid, in c.fetchall()])
		old_fids = [fid for fid in old_fids if "overwritten" not in self.get_fileresource(fid).info]
		
		###################################
		# Create and execute SQL commands #
		###################################
		sql_cmds = [] # A list of sql, parameters
		
		# Task
		sql_cmds.extend(_sql_insert_task(task))
		# Resource
		for resource in new_resources:
			sql_cmds.extend(_sql_insert_resource(resource))
		# FileResource
		for fileresource in new_fileresources:
			sql_cmds.extend(_sql_insert_fileresource(fileresource))
		
		# Overwrite files handling
		for fid in old_fids:
			sql_cmds.append(('INSERT INTO file_info(fid, info_key, info_value) VALUES(?,?,?)''', [fid, 'overwritten', end_time.isoformat()]))

		
		
		_sql_execute_commands(self.sql, sql_cmds)
		
		
		# Update internal db
		for resource in new_resources:
			self.resources_db[resource.rid] = resource
		for fileresource in new_fileresources:
			self.fileresources_db[fileresource.fid] = fileresource
		self.tasks_db[task.tid] = task
		
		# Fire RMS update event
		self.fireRMSUpdateEvent(list(map(lambda rmsobj: (RMSUpdateEvent.INSERT, rmsobj.get_full_id()), [task] + task.output_fileresources + task.output_resources)))
		
		# Return the task
		return task
		
	def _ba_rms_to_func_args_kwargs(self, ba):	
		return [self._obtain_rms_run_content(arg) for arg in ba.args], {k:self._obtain_rms_run_content(arg) for k, arg in ba.kwargs.items()}
	
	
	def run(self, pipe, args=(), kwargs={}, 
		task_description="", task_tags=set(), task_info={}, 
		resource_description="", resource_tags=set(), resource_info={},
		fileresource_description="", fileresource_tags=set(), fileresource_info={}):
		'''
		A normal way to run pipe with the given parameters.
		Register the run as a task, and register the output resources and fileresources.
		Return the resource or resources like normal function run.
		
		'''
		ba = _get_func_ba(pipe.func, args, kwargs)
		
		# Check for existing tasks. Avoid running tasks with exactly the same parameters
		if not self.force_rerun:
			prev_tasks = self.find_tasks_by_pipe_and_args(pipe.pid, ba.args, ba.kwargs)
			prev_tasks = [t for t in prev_tasks if "deprecated" not in t.info]
			if len(prev_tasks) > 0:
				return prev_tasks[0].return_values[0]
		
		# Conversion to func_args and func_kwargs
		func_args, func_kwargs = self._ba_rms_to_func_args_kwargs(ba)
		
		# Generate output files, this step ensures that the output func works 
		output_files = self.compute_output_files(pipe, func_args, func_kwargs)
		if not self.allow_overwrite:
			if self.exist_output_files(output_files):
				print(func_args, func_kwargs)
				raise Exception("File exist. Cannot overwrite files")

		# Add task info
		if self.scriptID is not None:
			task_info = {"scriptid":self.scriptID, **task_info}
		begin_time = datetime.datetime.now()
		raw_return_value = pipe.func(*func_args, **func_kwargs)
		end_time = datetime.datetime.now()
		
		task = self.register_finished_task(raw_return_value, begin_time, end_time, ba, pipe, 
										task_description, task_tags, task_info,
										resource_description, resource_tags, resource_info,
										fileresource_description, fileresource_tags, fileresource_info)
		
		return task.return_values[0]
		
	def _run_pipe(self, pipe, args=(), kwargs={}):
		'''
		Run pipe with the given parameters, but do not register anything
		Returns the normal return values of the pipe function.
		
		'''

		ba = _get_func_ba(pipe.func, args, kwargs)
		func_args, func_kwargs = self._ba_rms_to_func_args_kwargs(ba)
		return pipe.func(*func_args, **func_kwargs)

	def replace_virtualresource(self, virtualresource, resource):
		'''
		Replace a virtual resource with the desired resource.
		Can also replace a virtual filresource with the desired fileresource
		
		Should change it to use replace_unruntask_rmsobj 
		'''
		virtualresource.replacement = resource
		self.fireRMSUpdateEvent([(RMSUpdateEvent.REPLACE, virtualresource.get_full_id())])
		for unruntask in self.find_downstream_objs([virtualresource], 1):
			original_kv = list(unruntask.ba.arguments.items())
			for k, v in original_kv:
				if unruntask.ba.signature.parameters[k].kind == inspect.Parameter.VAR_POSITIONAL:
					unruntask.ba.arguments[k] = [resource if (isinstance(i, RMSEntry) and i is virtualresource) else i for i in v]
				elif unruntask.ba.signature.parameters[k].kind == inspect.Parameter.VAR_KEYWORD:
					unruntask.ba.arguments[k] = {ik:(resource if (isinstance(i, RMSEntry) and i is virtualresource) else i) for ik, i in v.items()}
				else:
					if isinstance(v, RMSEntry):
						if v is virtualresource:
							unruntask.ba.arguments[k] = resource
		#Temp disable
# 		self.delete(virtualresource.get_full_id())
		
	def replace_unruntask(self, unruntask, task):
		'''
		Update all associated virtual resources  
		'''
		if unruntask.return_values is not None:
			if len(task.return_values) == len(unruntask.return_values):
				for vr, r in zip(unruntask.return_values, task.return_values):
					self.replace_virtualresource(vr, r)
		if unruntask.output_files is not None:
			if len(task.output_files) == len(unruntask.output_files):
				for vr, r in zip(unruntask.output_files, task.output_files):
					self.replace_virtualresource(vr, r)
					
		
		unruntask.replacement = task
		self.fireRMSUpdateEvent([(RMSUpdateEvent.REPLACE, unruntask.get_full_id())])
		# Temp disable
		#self.delete(unruntask.get_full_id())

	def run_unruntask(self, unruntask):
		pipe = self.get_pipe(unruntask.pid)
		ba = unruntask.ba
		prev_tasks = self.find_tasks_by_pipe_and_args(pipe.pid, ba.args, ba.kwargs)
		if len(prev_tasks) > 0:
			print("The task has been done before.")
			task = prev_tasks[0]
		else:
			func_args, func_kwargs = self._ba_rms_to_func_args_kwargs(ba)
			
			begin_time = datetime.datetime.now()
			raw_return_value = pipe.func(*func_args, **func_kwargs)
			end_time = datetime.datetime.now()
		
			task = self.register_finished_task(raw_return_value, begin_time, end_time, ba, pipe, 
											unruntask.task_description, unruntask.task_tags, unruntask.task_info,
											unruntask.resource_description, unruntask.resource_tags, unruntask.resource_info,
											unruntask.fileresource_description, unruntask.fileresource_tags, unruntask.fileresource_info)
		
		self.replace_unruntask(unruntask, task)
		return task.return_values[0]
	
	def create_template_job(self, func, args, kwargs):
		tpid = self._get_new_id()
		return RMSTemplateJob(tpid, func, args, kwargs)
	
	def run_template_job(self, template_job):
		'''
		This involves running multiple steps in a function
		'''
		template_job.status = RMSTemplateJobStatus.RUNNING
# 		try:
		template_job.func(self, *template_job.args, **template_job.kwargs)
		template_job.status = RMSTemplateJobStatus.COMPLETE
		return template_job 
# 		except:
# 			print("Here")
# 			template_job.status = RMSTemplateJobStatus.ERROR
		
	def run_template(self, func, args=[], kwargs={}):
		job = self.create_template_job(func, args, kwargs)
		return self.run_template_job(job)
		
	def simulate_template(self, func, args=[], kwargs={}):
		simulator = TemplateSimulator()
		func(simulator, *args, **kwargs)
		return simulator.get_result()
	
	def run_unruntask_chain(self, unruntask):
		for vr in unruntask.input_virtualresources:
			prev_unruntasks = self.find_upstream_objs([vr], 1)
			if len(prev_unruntasks) > 1:
				assert False
			if len(prev_unruntasks) == 0:
				print("Cannot fetch virtual task: " + vr.get_id())
			prev_unruntask = next(iter(prev_unruntasks))
			self.run_unruntask_chain(prev_unruntask)
		if len(unruntask.input_virtualresources) > 0:
			raise ResourceNotReadyException()
		
		
		pipe = self.get_pipe(unruntask.pid)
		ba = unruntask.ba
		prev_tasks = self.find_tasks_by_pipe_and_args(pipe.pid, ba.args, ba.kwargs)
		if len(prev_tasks) > 0:
			print("The task has been done before.")
			task = prev_tasks[0]
		else:
			func_args, func_kwargs = self._ba_rms_to_func_args_kwargs(ba)
			
			begin_time = datetime.datetime.now()
			raw_return_value = pipe.func(*func_args, **func_kwargs)
			end_time = datetime.datetime.now()
		
			task = self.register_finished_task(raw_return_value, begin_time, end_time, ba, pipe, 
											unruntask.task_description, unruntask.task_tags, unruntask.task_info,
											unruntask.resource_description, unruntask.resource_tags, unruntask.resource_info,
											unruntask.fileresource_description, unruntask.fileresource_tags, unruntask.fileresource_info)
		
		self.replace_unruntask(unruntask, task)

		return task.return_values[0]
		
	
	def create_unruntask(self, pid, args=[], kwargs={}, return_values=[], output_files=[],
						task_description="", task_tags=set(), task_info={},
						resource_description="", resource_tags=set(), resource_info={}, 
						fileresource_description="", fileresource_tags=set(), fileresource_info={}
						):
		uid = self._get_new_id()
		pipe = self.get_pipe(pid)
		ba = _get_func_ba(pipe.func, args, kwargs, partial=True)
		ut = UnrunTask(uid, pid, ba, return_values, output_files,
					task_description, task_tags, task_info,
					resource_description, resource_tags, resource_info,
					fileresource_description, fileresource_tags, fileresource_info
					)
		
		for vr in return_values:
			vr.uid = uid
		if output_files is not None:
			for vr in output_files:
				vr.uid = uid
		self.unruntasks_db[uid] = ut
		self.fireRMSUpdateEvent([(RMSUpdateEvent.INSERT, ut.get_full_id())])
		return ut
	
	def create_unruntask_from_task(self, task, return_values=[], output_files=[]):
		uid = self._get_new_id()
		ba = _get_func_ba(self.get_pipe(task.pid).func, task.args, task.kwargs, partial=True)
		unruntask = UnrunTask(uid, task.pid, ba, return_values, output_files)
		self.unruntasks_db[uid] = unruntask
		return unruntask
	def create_virtualresource(self):
		vid = self._get_new_id()
		vr = VirtualResource(vid)
		self.virtualresources_db[vid] = vr
		self.fireRMSUpdateEvent([(RMSUpdateEvent.INSERT, vr.get_full_id())])
		return vr
	def create_virtualfileresource(self, file_path):
		vid = self._get_new_id()
		vr = VirtualResource(vid, file_path)
		self.virtualresources_db[vid] = vr
		self.fireRMSUpdateEvent([(RMSUpdateEvent.INSERT, vr.get_full_id())])
		return vr
	def create_unruntask_chain(self, args):
		rmsobjs = list(map(self._as_rmsobj, args))
		
		newobjs = []
		
		g = self.construct_digraph_from_rmsids(rmsobjs)
		virtual_rmsobjs_dict = {} 
		for rmsid in nx.topological_sort(g):
			rmsobj = self.get(rmsid)
			if rmsid[0] == RMSEntryType.RESOURCE:
				vr = self.create_virtualresource()
				virtual_rmsobjs_dict[rmsobj] = vr
			elif rmsid[0] == RMSEntryType.FILERESOURCE:
				vr = self.create_virtualresource()
				virtual_rmsobjs_dict[rmsobj] = vr
			else:
				pass
		newobjs.extend(virtual_rmsobjs_dict.values())
		
		
		for rmsid in nx.topological_sort(g):
			if rmsid[0] == RMSEntryType.TASK:
				rmsobj = self.get(rmsid)
				# Update return values and output_files
				return_values = []
				output_files = []
				for r in rmsobj.return_values:
					if r in virtual_rmsobjs_dict:
						return_values.append(virtual_rmsobjs_dict[r])
					else:
						vr = self.create_virtualresource()
						newobjs.append(vr)
						return_values.append(vr)
				for r in rmsobj.output_files:
					if r in virtual_rmsobjs_dict:
						output_files.append(virtual_rmsobjs_dict[r])
					else:
						vr = self.create_virtualresource()
						newobjs.append(vr)
						output_files.append(vr)
				
				unruntask = self.create_unruntask_from_task(rmsobj, output_files=output_files, return_values=return_values)
				original_kv = list(unruntask.ba.arguments.items())
				for k, v in original_kv:
					if unruntask.ba.signature.parameters[k].kind == inspect.Parameter.VAR_POSITIONAL:
						unruntask.ba.arguments[k] = [virtual_rmsobjs_dict[i] if (isinstance(i, RMSEntry) and i in virtual_rmsobjs_dict) else i for i in v]
					elif unruntask.ba.signature.parameters[k].kind == inspect.Parameter.VAR_KEYWORD:
						unruntask.ba.arguments[k] = {ik:(virtual_rmsobjs_dict[i] if (isinstance(i, RMSEntry) and i in virtual_rmsobjs_dict) else i) for ik, i in v.items()}
					else:
						if isinstance(v, RMSEntry):
							if v in virtual_rmsobjs_dict:
								unruntask.ba.arguments[k] = virtual_rmsobjs_dict[v]
								
				newobjs.append(unruntask)
			else:
				pass
		
		self.fireRMSUpdateEvent([(RMSUpdateEvent.INSERT, rmsobj.get_full_id()) for rmsobj in newobjs])
		return newobjs
					
	def _guess_unruntask_output(self, unruntask):
		'''
		Based on the current parameter of unruntask, guess the output of the unruntask by adding default or None to missing parameters
		'''
		if self.get_pipe(unruntask.pid).output_func is None:
			return []
		ba = unruntask.ba.signature.bind_partial(*unruntask.ba.args, **unruntask.ba.kwargs)
		for p, param in unruntask.ba.signature.parameters.items():
			if param.kind == inspect.Parameter.VAR_POSITIONAL or param.kind == inspect.Parameter.VAR_KEYWORD:
				continue
			if p not in ba.arguments:
				if param.default is param.empty:
					ba.arguments[p] = None
				else:
					ba.arguments[p] = param.default
		try:
			potential_output_files = self.get_pipe(unruntask.pid).output_func(*ba.args, **ba.kwargs)
			return potential_output_files
		except:
			return []
		
	def mark_deprecated(self, rmsids, mark_downstream=True):
		rmsids = [self._as_rmsid(rmsid) for rmsid in rmsids]
		rmsids = set(rmsids)
		if mark_downstream:
			rmsids = set.union(rmsids, self.find_downstream_rmsids(rmsids))
		self._prefetch_entries(rmsids=rmsids)
		rmsids = [rmsid for rmsid in rmsids if "deprecated" not in self.get(rmsid).info]
		cur_time_str = datetime.datetime.now().isoformat()
		sql_cmds = []
		g = _group_by_first_id(rmsids)
		fids = [i[1] for i in g[RMSEntryType.FILERESOURCE]]
		sql_cmds.append(['INSERT INTO file_info(fid, info_key, info_value) VALUES ''' + ",".join(["(?,?,?)"] * len(fids)), list(itertools.chain.from_iterable([[fid, 'deprecated', cur_time_str] for fid in fids]))])
		rids = [i[1] for i in g[RMSEntryType.RESOURCE]]
		sql_cmds.append(['INSERT INTO resource_info(rid, info_key, info_value) VALUES ''' + ",".join(["(?,?,?)"] * len(rids)), list(itertools.chain.from_iterable([[rid, 'deprecated', cur_time_str] for rid in rids]))])
		tids = [i[1] for i in g[RMSEntryType.TASK]]
		sql_cmds.append(['INSERT INTO task_info(tid, info_key, info_value) VALUES ''' + ",".join(["(?,?,?)"] * len(tids)), list(itertools.chain.from_iterable([[tid, 'deprecated', cur_time_str] for tid in tids]))])
		_sql_execute_commands(self.sql, sql_cmds)
		self._prefetch_entries(rmsids=rmsids, refetch=True)
		self.fireRMSUpdateEvent([(RMSUpdateEvent.MODIFY, full_id) for full_id in rmsids])
	
	def regenerate_missing_files(self, args, require_no_existing_files=True):
		rmsids = []
		for arg in args:
			if isinstance(arg, str):
				rmsids.append(self.file_from_path(arg).get_full_id())
			else:
				rmsids.append(arg)
		corresponding_task_rmsids = self.find_upstream_rmsids(rmsids, distance=1)
		self._prefetch_entries(rmsids=corresponding_task_rmsids)
		
		to_regenerate_file_rmsids = set([f.get_full_id() for rmsid in corresponding_task_rmsids for f in self.get(rmsid).output_fileresources])
		if not set(rmsids).issubset(to_regenerate_file_rmsids):
			raise Exception("Some files cannot be regenerated.")
		
		
		rmsids_with_existing_files = [rmsid for rmsid in to_regenerate_file_rmsids if os.path.exists(self.get(rmsid).file_path)]
		if len(rmsids_with_existing_files) > 0:
			fs = ";".join([self.get(rmsid).file_path for rmsid in rmsids_with_existing_files])
			if require_no_existing_files:
				raise Exception("Some files exist! Process halts and no existing files are overwritten. " + fs)
			else:
				print("Warning: Some files exist! "+ fs)
		for rmsid in corresponding_task_rmsids:
			task = self.get(rmsid)
			pipe = self.get_pipe(task.pid)
			raw_return_value = self._run_pipe(pipe, task.args, task.kwargs)
			for f in task.output_fileresources:
				md5 = _get_file_md5(f.file_path)
				if md5 != f.md5:
					print(f"Warning: newly regenerated file has different md5: Old - {f.md5}; New - {md5}")
			
			

		
	##########################
	# Users edit an RMSEntry #
	##########################
	def update(self, full_id, **kwargs):
		'''
		This methods directly update the RMSEntry by field name in the database
		'''
		# Not designed for modifying IDs
		sql_cmds = [] 
		sql_cmds.append(_sql_update(full_id, **kwargs))
		_sql_execute_commands(self.sql, sql_cmds)
		new_rmsobj = self.get(full_id, refetch=True) # Refetch is required to update the copy in the memory
		self.fireRMSUpdateEvent([(RMSUpdateEvent.MODIFY, full_id)])
		return new_rmsobj
	
	def update_unruntask_arguments(self, full_id, updated_ba_arguments):
		self.get(full_id).ba.arguments.update(updated_ba_arguments)
		self.fireRMSUpdateEvent([(RMSUpdateEvent.MODIFY, full_id)])

	###########################
	# Users delete RMSEntries #
	###########################
	
	def _get_depenedent_rmsids(self, rmsids):
		dependent_rmsids = set()
		for rmsid in rmsids:
			rmsid = self._as_rmsid(rmsid)
			if rmsid[0] == RMSEntryType.RESOURCE:
				dependent_rmsids.update([task.get_full_id() for task in self.find_tasks_by_io("io", rids=[rmsid[1]])])
			elif rmsid[0] == RMSEntryType.FILERESOURCE:
				dependent_rmsids.update([task.get_full_id() for task in self.find_tasks_by_io("io", fids=[rmsid[1]])])
			elif rmsid[0] == RMSEntryType.TASK:
				dependent_rmsids.update([rmsobj.get_full_id() for rmsobj in self.get_task(rmsid[1]).output_fileresources + self.get_task(rmsid[1]).output_resources])
			elif rmsid[0] == RMSEntryType.PIPE:
				dependent_rmsids.update([task.get_full_id() for task in self.find_tasks_by_pipe([rmsid[1]])])
				dependent_rmsids.update([task.get_full_id() for task in self.find_tasks_by_io("io", pids=[rmsid[1]])])
# 			elif rmsid[0] == RMSEntryType.UNRUNTASK:
# 				pass
# 			elif rmsid[0] == RMSEntryType.VIRTUALRESOURCE:
# 				pass
			else:
				raise Exception("The type is not supported: " + str(rmsid[0]))
		return dependent_rmsids
	def _get_delete_sql_cmds(self, rmsids):
		
		sql_cmds = []
		for rmsid in rmsids:
			if rmsid[0] == RMSEntryType.RESOURCE:
				sql_cmds.append([f"DELETE FROM resources WHERE rid=?", [rmsid[1]]])
				sql_cmds.append([f"DELETE FROM resource_info WHERE rid=?", [rmsid[1]]])
				sql_cmds.append([f"DELETE FROM resource_tags WHERE rid=?", [rmsid[1]]])
			elif rmsid[0] == RMSEntryType.FILERESOURCE:
				sql_cmds.append([f"DELETE FROM files WHERE fid=?", [rmsid[1]]])
				sql_cmds.append([f"DELETE FROM file_info WHERE fid=?", [rmsid[1]]])
				sql_cmds.append([f"DELETE FROM file_tags WHERE fid=?", [rmsid[1]]])
	
			elif rmsid[0] == RMSEntryType.TASK:
				sql_cmds.append([f"DELETE FROM tasks WHERE tid=?", [rmsid[1]]])
				sql_cmds.append([f"DELETE FROM tasks_args_json WHERE tid=?", [rmsid[1]]])
				sql_cmds.append([f"DELETE FROM tasks_args_pipe WHERE tid=?", [rmsid[1]]])
				sql_cmds.append([f"DELETE FROM tasks_args_file WHERE tid=?", [rmsid[1]]])
				sql_cmds.append([f"DELETE FROM tasks_args_resource WHERE tid=?", [rmsid[1]]])
				sql_cmds.append([f"DELETE FROM tasks_kwargs_json WHERE tid=?", [rmsid[1]]])
				sql_cmds.append([f"DELETE FROM tasks_kwargs_pipe WHERE tid=?", [rmsid[1]]])
				sql_cmds.append([f"DELETE FROM tasks_kwargs_file WHERE tid=?", [rmsid[1]]])
				sql_cmds.append([f"DELETE FROM tasks_kwargs_resource WHERE tid=?", [rmsid[1]]])
				sql_cmds.append([f"DELETE FROM tasks_returnvalue WHERE tid=?", [rmsid[1]]])
				sql_cmds.append([f"DELETE FROM tasks_outputfiles WHERE tid=?", [rmsid[1]]])
				sql_cmds.append([f"DELETE FROM task_info WHERE tid=?", [rmsid[1]]])
				sql_cmds.append([f"DELETE FROM task_tags WHERE tid=?", [rmsid[1]]])
			elif rmsid[0] == RMSEntryType.PIPE:
				sql_cmds.append([f"DELETE FROM pipes WHERE pid=?", [rmsid[1]]])
				sql_cmds.append([f"DELETE FROM pipe_info WHERE pid=?", [rmsid[1]]])
				sql_cmds.append([f"DELETE FROM pipe_tags WHERE pid=?", [rmsid[1]]])
			else:
				raise Exception()

		return sql_cmds
	def delete(self, *rmsids):
		'''
		Delete an entry. Use it with care as the action cannot be undone.
		
		The delete function first looks for all dependencies of the current RMS objects to be deleted. 
		If any of the RMS object dependency breaks, an exception is raised
		
		Support for virtual RMS objects (UNRUNTASK / VIRTUALRESOURCE) may be problematic.
		'''
		virtual_rmsids = [rmsid for rmsid in rmsids if rmsid[0] == RMSEntryType.UNRUNTASK or rmsid[0] == RMSEntryType.VIRTUALRESOURCE]
		rmsids = [rmsid for rmsid in rmsids if not (rmsid[0] == RMSEntryType.UNRUNTASK or rmsid[0] == RMSEntryType.VIRTUALRESOURCE)]	
				
		rmsids = set(rmsids)
		dependent_rmsids = self._get_depenedent_rmsids(rmsids)
		if not dependent_rmsids.issubset(rmsids):
			raise Exception("Dependency of the following rmsids will break:\n" + "\n".join(["- " + str(rmsid) for rmsid in (dependent_rmsids - rmsids)]) + "\nwhere deleted rmsids are:\n" + "\n".join(["- " + str(rmsid) for rmsid in rmsids]))
		
		
		sql_cmds = []
		sql_cmds.extend(self._get_delete_sql_cmds(rmsids))
		_sql_execute_commands(self.sql, sql_cmds)
		for rmsid in rmsids:
			try: 
				if rmsid[0] == RMSEntryType.RESOURCE:
					self.resources_db.pop(rmsid[1])
				elif rmsid[0] == RMSEntryType.FILERESOURCE:
					self.fileresources_db.pop(rmsid[1])
				elif rmsid[0] == RMSEntryType.TASK:
					self.tasks_db.pop(rmsid[1])
				elif rmsid[0] == RMSEntryType.PIPE:
					self.pipes_db.pop(rmsid[1])
				else:
					raise Exception()
			except KeyError:
				pass
		for rmsid in virtual_rmsids:
			if rmsid[0] == RMSEntryType.UNRUNTASK:
				self.unruntasks_db.pop(rmsid[1])
			elif rmsid[0] == RMSEntryType.VIRTUALRESOURCE:
				self.virtualresources_db.pop(rmsid[1])
			else:
				raise Exception()
		self.fireRMSUpdateEvent(list(map(lambda rmsid: (RMSUpdateEvent.DELETE, rmsid), list(rmsids) + list(virtual_rmsids))))
		
		
	####################################################
	# Import
	####################################################
			
		
	def rmsimport(self, namespace=None, moduleobj=None, varname = None, altname=None, return_first_module=False):
		'''
		Perform a fake import
	
		'''
		def is_bound_method(func):
			return inspect.ismethod(func)
		pipe_results = self.sql.cursor().execute(f"SELECT pid FROM pipes;").fetchall()
		pipes = [self.get_pipe(pid) for pid, in pipe_results]
		
		if namespace is None:
			namespace = dict()
		if moduleobj is None:
			moduleobj = "builtins"

		# There are two ways to add bound_method. 
		# 1. If the class is already registered, set the bound method as the class attr
		# 2. If the class is not registered, set the class as another virtual module
		bound_method_classes = defaultdict(list)
		for p in pipes:
			if inspect.ismethod(p.func):
				if inspect.isclass(p.func.__self__):
					bound_method_classes[p.func.__self__].append(p)
		
	
		# Import the module
		if varname is None:
			output_varnames = []
			output_pipes = []
			for pipe in pipes:
				if pipe.func.__module__ == moduleobj and not inspect.ismethod(pipe.func): #(inspect.isclass(pipe.func) or inspect.isfunction(pipe.func)):
					output_varnames.append(pipe.func.__name__)
					output_pipes.append(pipe)
			d = {v:p for v, p in zip(output_varnames, output_pipes)}
			if altname is None:
				vm = self.vm
				for i in moduleobj.split("."):
					if not hasattr(vm, i):
						setattr(vm, i, VirtualModule())
					vm = getattr(vm, i)
				for v, p in d.items():
					setattr(vm, v, p)
				
				# Deal with bound method that is not in output pipes
				for k, bound_methods in bound_method_classes.items():
					if k.__module__ in moduleobj:
						for p in output_pipes:
							if k is p.func:
								break
						else:
							i = k.__name__
							if not hasattr(vm, i):
								setattr(vm, i, VirtualModule())
							for bound_method in bound_methods:
								setattr(getattr(vm, i), bound_method.func.__name__, bound_method)
				first_vm_name = moduleobj.split(".")[0]
				namespace[first_vm_name] = getattr(self.vm, first_vm_name)
				if return_first_module:
					returnvalue = namespace[first_vm_name]
				else:
					returnvalue = vm
			else:
				# Deal with bound method that is not in output pipes
				to_add = {}
				for k, bound_methods in bound_method_classes.items():
					if k.__module__ in moduleobj:
						for p in output_pipes:
							if k is p.func:
								break
						else:
							i = k.__name__
							if i not in to_add:
								to_add[i] = VirtualModule()
							for bound_method in bound_methods:
								setattr(to_add[i], bound_method.func.__name__, bound_method)
				namespace[altname] = types.SimpleNamespace(**d, **to_add)
				returnvalue = namespace[altname]
				
		# Import the variable
		else:
			output_pipes = []
			# wild card case
			if varname == "*":
				tmp_returnvalues = {}
				for pipe in pipes:
					if pipe.func.__module__ == moduleobj and not inspect.ismethod(pipe.func):#(inspect.isclass(pipe.func) or inspect.isfunction(pipe.func)):
						namespace[pipe.func.__name__] = pipe
						output_pipes.append(pipe)
						tmp_returnvalues[pipe.func.__name__] = namespace[pipe.func.__name__]
				to_add = {}
				for k, bound_methods in bound_method_classes.items():
					if k.__module__ in moduleobj:
						for p in output_pipes:
							if k is p.func:
								break
						else:
							i = k.__name__
							if i not in to_add:
								to_add[i] = VirtualModule()
							for bound_method in bound_methods:
								setattr(to_add[i], bound_method.func.__name__, bound_method)
				for k, p in to_add.items():
					namespace[k] = p
					tmp_returnvalues[k] = namespace[k]
				if len(tmp_returnvalues) < 1:
					raise Exception(f"Cannot find anything to import in {moduleobj}")
				returnvalue = tmp_returnvalues
				
			# normal case
			else:
				tmp_returnvalues = []
				for pipe in pipes:
					if pipe.func.__module__ == moduleobj and not inspect.ismethod(pipe.func):
						if pipe.func.__name__ == varname:
							if altname is None:
								altname = varname
							namespace[altname] = pipe
							output_pipes.append(pipe)
							tmp_returnvalues.append(pipe)
				to_add = {}
				
				for k, bound_methods in bound_method_classes.items():
					if k.__name__ == varname and k.__module__ in moduleobj:
						for p in output_pipes:
							if k is p.func:
								break
						else:
							i = k.__name__
							if i not in to_add:
								to_add[i] = VirtualModule()
							for bound_method in bound_methods:
								setattr(to_add[i], bound_method.func.__name__, bound_method)
				for k, p in to_add.items():
					namespace[k] = p
					tmp_returnvalues.append(namespace[k])
				if len(tmp_returnvalues) < 1:
					raise Exception(f"Cannot find {varname} in {moduleobj}")
				if len(tmp_returnvalues) > 1:
					assert False
				returnvalue = tmp_returnvalues[0]
				
		for pipe in output_pipes:
			if inspect.isclass(pipe.func):
				if pipe.func in bound_method_classes:
					for bound_method in bound_method_classes[pipe.func]:
						setattr(pipe, bound_method.func.__name__, bound_method)
							
		return returnvalue
	########################
	# Users exports RMS DB #
	########################
	def export_db_snapshot(self, output_dbview, rmsobjs):
		'''
		Create a db view snapshot
		Make sure all rmsobjs depenedencies MUST be resolved. 
		For example, if you export any task, you must also export the corresponding pipe, inputs and outputs.
		'''
		# Check if output file exists
		# Since there is a high chance some people may overwrite their own database, auto-overwrite is disabled
		if os.path.exists(output_dbview):
			raise Exception("File already exists. Auto-overwrite is disabled to avoid overwriting the original DB.")
		sql = sqlite3.connect(output_dbview)
		
		# Create the initial database
		cmds = [
"""CREATE TABLE IF NOT EXISTS metainfo (
	infokey text NOT NULL,
	infovalue text NOT NULL
);
""",
"""
CREATE TABLE IF NOT EXISTS tasks (
	tid text PRIMARY KEY,
	pid text NOT NULL,
	begin_time text NOT NULL, 
	end_time text NOT NULL,
	description text,
	FOREIGN KEY (pid) REFERENCES pipes(pid)
);
""",
"""
CREATE TABLE IF NOT EXISTS tasks_args_json (
	tid text NOT NULL,
	arg_order integer NOT NULL,
	arg_value text NOT NULL,
	FOREIGN KEY (tid) REFERENCES tasks(tid)
);
""",
"""
CREATE TABLE IF NOT EXISTS tasks_args_resource (
	tid text NOT NULL,
	arg_order integer NOT NULL,
	rid text NOT NULL,
	FOREIGN KEY (tid) REFERENCES tasks(tid),
	FOREIGN KEY (rid) REFERENCES resources(rid)
);
""",
"""
CREATE TABLE IF NOT EXISTS tasks_args_file (
	tid text NOT NULL,
	arg_order integer NOT NULL,
	fid text NOT NULL,
	FOREIGN KEY (tid) REFERENCES tasks(tid),
	FOREIGN KEY (fid) REFERENCES files(fid)
);
""",
"""
CREATE TABLE IF NOT EXISTS tasks_args_pipe (
	tid text NOT NULL,
	arg_order integer NOT NULL,
	pid text NOT NULL,
	FOREIGN KEY (tid) REFERENCES tasks(tid),
	FOREIGN KEY (pid) REFERENCES pipes(pid)
);
""",
"""
CREATE TABLE IF NOT EXISTS tasks_kwargs_json (
	tid text NOT NULL,
	arg_key text NOT NULL,
	arg_value text NOT NULL,
	FOREIGN KEY (tid) REFERENCES tasks(tid)
);
""",
"""
CREATE TABLE IF NOT EXISTS tasks_kwargs_resource (
	tid text NOT NULL,
	arg_key text NOT NULL,
	rid text NOT NULL,
	FOREIGN KEY (tid) REFERENCES tasks(tid),
	FOREIGN KEY (rid) REFERENCES resources(rid)
);
""",
"""
CREATE TABLE IF NOT EXISTS tasks_kwargs_file (
	tid text NOT NULL,
	arg_key text NOT NULL,
	fid text NOT NULL,
	FOREIGN KEY (tid) REFERENCES tasks(tid),
	FOREIGN KEY (fid) REFERENCES files(fid)
);
""",
"""
CREATE TABLE IF NOT EXISTS tasks_kwargs_pipe (
	tid text NOT NULL,
	arg_key text NOT NULL,
	pid text NOT NULL,
	FOREIGN KEY (tid) REFERENCES tasks(tid),
	FOREIGN KEY (pid) REFERENCES pipes(pid)
);
""",
"""CREATE TABLE IF NOT EXISTS tasks_returnvalue (
	tid text NOT NULL,
	rid text NOT NULL,
	FOREIGN KEY (tid) REFERENCES tasks(tid),
	FOREIGN KEY (rid) REFERENCES resources(rid)
);
""",
"""CREATE TABLE IF NOT EXISTS tasks_outputfiles (
	tid text NOT NULL,
	forder int NOT NULL,
	fid text NOT NULL,
	FOREIGN KEY (tid) REFERENCES tasks(tid),
	FOREIGN KEY (fid) REFERENCES files(fid)
);
""",
"""
CREATE TABLE IF NOT EXISTS pipes (
	pid text PRIMARY KEY,
	func text NOT NULL,
	return_volatile integer NOT NULL,
	is_deterministic integer NOT NULL,
	module_name text,
	func_name text,
	output_func text,
	description text
);
""",
"""
CREATE TABLE IF NOT EXISTS resources (
	rid text PRIMARY KEY,
	volatile integer,
	description text
);
""",
"""
CREATE TABLE IF NOT EXISTS files (
	fid text PRIMARY KEY,
	file_path text,
	md5 text,
	description text
);
""",
"""
CREATE TABLE IF NOT EXISTS task_tags (
	tid text NOT NULL,
	tag_value text,
	FOREIGN KEY (tid) REFERENCES tasks(tid)
);
""",
"""
CREATE TABLE IF NOT EXISTS task_info (
	tid text NOT NULL,
	info_key text,
	info_value text,
	FOREIGN KEY (tid) REFERENCES tasks(tid)
);
""",
"""
CREATE TABLE IF NOT EXISTS resource_tags (
	rid text NOT NULL,
	tag_value text,
	FOREIGN KEY (rid) REFERENCES resources(rid)
);
""",
"""
CREATE TABLE IF NOT EXISTS resource_info (
	rid text NOT NULL,
	info_key text,
	info_value text,
	FOREIGN KEY (rid) REFERENCES resources(rid)
);
""",
"""
CREATE TABLE IF NOT EXISTS pipe_tags (
	pid text NOT NULL,
	tag_value text,
	FOREIGN KEY (pid) REFERENCES pipes(pid)
);
""",
"""
CREATE TABLE IF NOT EXISTS pipe_info (
	pid text NOT NULL,
	info_key text,
	info_value text,
	FOREIGN KEY (pid) REFERENCES pipes(pid)
	
);
""",
"""
CREATE TABLE IF NOT EXISTS file_tags (
	fid text NOT NULL,
	tag_value text,
	FOREIGN KEY (fid) REFERENCES files(fid)
);
""",
"""
CREATE TABLE IF NOT EXISTS file_info (
	fid text NOT NULL,
	info_key text,
	info_value text,
	FOREIGN KEY (fid) REFERENCES files(fid)
	
);
""",
]
		c = sql.cursor()
		c.execute("BEGIN")
		try:
			for cmd in cmds:
				c.execute(cmd)
			c.execute("COMMIT")
		except sql.Error as e:
			print("Encounter error when creating DB")
			print(e)
			c.execute("ROLLBACK")
		
		if len(c.execute('SELECT * FROM metainfo WHERE infokey = ?', ['dbid']).fetchall()) != 0:
			raise Exception()
		c.execute("BEGIN")
		c.execute('INSERT INTO metainfo(infokey, infovalue) VALUES(?, ?)', ["dbid", self.database_id])
		c.execute('INSERT INTO metainfo(infokey, infovalue) VALUES(?, ?)', ["db_snapshot_time", datetime.datetime.now().isoformat()])
		
		c.execute("COMMIT")

		
		
		split_rmsids = defaultdict(list)
		for rmsobj in rmsobjs:
			rmsid = self._as_rmsid(rmsobj)
			split_rmsids[rmsid[0]].append(rmsid[1])

		# Grab the entries and inserts
		conversion_dict = {"fid":RMSEntryType.FILERESOURCE, 
		"pid":RMSEntryType.PIPE, 
		"rid":RMSEntryType.RESOURCE, 
		"tid":RMSEntryType.TASK}
		template_cmds= ['INSERT INTO files(fid, file_path, md5, description) VALUES(?,?,?,?)',
		'INSERT INTO file_tags(fid, tag_value) VALUES(?,?)',
		'INSERT INTO file_info(fid, info_key, info_value) VALUES(?,?,?)',
		'INSERT INTO tasks_args_resource(tid, arg_order, rid) VALUES(?,?,?)',
		'INSERT INTO tasks_args_file(tid, arg_order, fid) VALUES(?,?,?)',
		'INSERT INTO tasks_args_pipe(tid, arg_order, pid) VALUES(?,?,?)',
		'INSERT INTO tasks_args_json(tid, arg_order, arg_value) VALUES(?,?,?)',
		'INSERT INTO tasks_kwargs_resource(tid, arg_key, rid) VALUES(?,?,?)',
		'INSERT INTO tasks_kwargs_file(tid, arg_key, fid) VALUES(?,?,?)',
		'INSERT INTO tasks_kwargs_pipe(tid, arg_key, pid) VALUES(?,?,?)',
		'INSERT INTO tasks_kwargs_json(tid, arg_key, arg_value) VALUES(?,?,?)',
		'INSERT INTO tasks_returnvalue(tid, rid) VALUES(?,?)',
		'INSERT INTO tasks_outputfiles(tid, forder, fid) VALUES(?,?,?)',
		'INSERT INTO tasks(tid, pid, begin_time, end_time, description) VALUES(?,?,?,?,?)' ,
		'INSERT INTO task_tags(tid, tag_value) VALUES(?,?)',
		'INSERT INTO task_info(tid, info_key, info_value) VALUES(?,?,?)',
		'INSERT INTO pipes(pid, func, return_volatile, is_deterministic, module_name, func_name, output_func, description) VALUES(?,?,?,?,?,?,?,?)' ,
		'INSERT INTO pipe_tags(pid, tag_value) VALUES(?,?)',
		'INSERT INTO pipe_info(pid, info_key, info_value) VALUES(?,?,?)',
		'INSERT INTO resources(rid, volatile, description) VALUES(?,?,?)',
		'INSERT INTO resource_tags(rid, tag_value) VALUES(?,?)',
		'INSERT INTO resource_info(rid, info_key, info_value) VALUES(?,?,?)'
		]
		cmds = []
		import re
		rms_cursor = self.sql.cursor()
		for template_cmd in template_cmds:
			tblname, target = re.match(r"^INSERT INTO ([^(]+)\(([^,]+),.+$", template_cmd).groups()
			target_rmsids = split_rmsids[conversion_dict[target]]
			records = rms_cursor.execute(f'SELECT * FROM {tblname} WHERE {target} IN ({",".join("?" * len(target_rmsids))})', target_rmsids).fetchall()
			for r in records:
				cmds.append((template_cmd, r))
		_sql_execute_commands(sql, cmds)
	def export_db_snapshot_for_files(self, output_dbview, target_files, source_files=[]):
		'''
		Create DB snapshots that include the entire process to generate the target files.
		If source files are provided, upstream of source files are not included.
		'''
		source_ids = [self.file_from_path(t).get_full_id() for t in source_files]
		result_ids = [self.file_from_path(t).get_full_id() for t in target_files]
		all_rmsobjs = list(self.find_upstream_objs(
			result_ids,
			rmsobj_continue_search_criteria=lambda rmsobj: rmsobj.get_full_id() not in source_ids
		)) + result_ids
		to_add = []
		for arg in all_rmsobjs:
			arg = self._as_rmsobj(arg)
			if arg.get_type() == RMSEntryType.TASK:
				to_add.extend(arg.output_fileresources + arg.output_resources)
				to_add.append((RMSEntryType.PIPE, arg.pid))
		all_rmsobjs.extend(to_add)
		to_export_rmsids = set(self._as_rmsid(i) for i in all_rmsobjs)
		self.export_db_snapshot(
			output_dbview, to_export_rmsids
		)		
		
	def construct_digraph_from_rmsids(self, rmsobjs):
		rmsobjs = list(map(self._as_rmsobj, rmsobjs))
		g = nx.DiGraph()
		g.add_nodes_from(map(lambda rmsobj: rmsobj.get_full_id(), rmsobjs))
		for rmsobj in rmsobjs:
			# instead of finding downstream, finding upstream is much more efficient
			for upstream_rmsobj in self.find_upstream_objs([rmsobj], 1):
				if upstream_rmsobj.get_full_id() in g:
					g.add_edge(upstream_rmsobj.get_full_id(), rmsobj.get_full_id())
		return g
	
	def reconstruct_digraph(self, rids=[], fids=[], tids=[]):
		'''
		Reconstructs the pipeline that yield the rids and fids
		
		'''
		g = nx.DiGraph()
		all_rids = set(); all_rids.update(rids)
		all_fids = set(); all_fids.update(fids)
		all_tids = set(); all_tids.update(tids)
		resources = [self.get_resource(rid) for rid in rids]
		fileresources = [self.get_fileresource(fid) for fid in fids]
		tasks = [self.get_task(tid) for tid in tids]
		while len(resources) > 0 or len(fileresources) > 0 or len(tasks) > 0:
			if len(tasks) > 0:
				task = tasks.pop()
				g.add_node((task.get_type(), task.get_id()))
				for resource in task.input_resources:
					g.add_edge((resource.get_type(), resource.get_id()), (task.get_type(), task.get_id()))
					if resource.rid not in all_rids:
						resources.append(resource)
						all_rids.add(resource.rid)
				for fileresource in task.input_fileresources:
					g.add_edge((fileresource.get_type(), fileresource.get_id()), (task.get_type(), task.get_id()))
					if fileresource.fid not in all_fids:
						fileresources.append(fileresource)
						all_fids.add(fileresource.fid)
			else:
				if len(resources) > 0:
					resource = resources.pop()
					if resource.tid is None:
						continue
					task = self.get_task(resource.tid)
					cid = (resource.get_type(), resource.get_id())
				elif len(fileresources) > 0:
					fileresource = fileresources.pop()
					if fileresource.tid is None:
						continue
					task = self.get_task(fileresource.tid)
					cid = (fileresource.get_type(), fileresource.get_id())
				else:
					raise Exception()
				g.add_node(cid)
				g.add_edge((task.get_type(), task.get_id()), cid)
				if task.tid not in all_tids:
					tasks.append(task)
					all_tids.add(task.tid)
		return g

	###
	# Currently the codes conversion to normal codes, RMS codes, function are all based on similar mechanism.
	# I need to edit it   
	# 
	###


		
	def convert_to_normal_codes(self, rids=[], fids=[], tids=[]):
		import textwrap
		
		g = self.reconstruct_digraph(rids, fids, tids)
		graph_node_mapping = {}
		
		next_dummy_var_id = 1 
		next_dummy_func_id = 1
		rid_var_dict = {}
		funcs = dict()
		redefined_pipes = {}
		def obtain_pipename(pipe):
			nonlocal next_dummy_func_id
			if pipe.func.__module__ == "__main__":
				if pipe.pid not in redefined_pipes:
					if pipe.func.__name__ == "<lambda>":
						new_name = "f"
					else:
						new_name = pipe.func.__name__
					while new_name in funcs:
						new_name = new_name + str(next_dummy_func_id)
						next_dummy_func_id += 1
					funcs[new_name] = (dill.dumps(pipe.func).hex(), pipe.info["sourcecode"] if "sourcecode" in pipe.info else None)
					redefined_pipes[pipe.pid] = new_name
				pipename = redefined_pipes[pipe.pid]
			elif pipe.func.__module__ == "builtins":
				pipename = pipe.func.__name__
			else:
				modules.add(pipe.func.__module__)
				pipename = pipe.func.__module__ + "." + pipe.func.__name__
			return pipename
		
		def convert_arg(arg):
			if isinstance(arg, Resource):
				return rid_var_dict[arg.rid]
			elif isinstance(arg, FileResource):
				return json.dumps(arg.file_path)
			elif isinstance(arg, Pipe):
				#return f'rmsp.get_pipe("{arg.pid}")'
				return obtain_pipename(arg)
			else:
				return repr(arg)
				#return json.dumps(arg)
		codes = []
		modules = set()
		for node in nx.topological_sort(g):
			if node[0] == RMSEntryType.TASK:
				task = self.get_task(node[1])
				# Add description as comment
				if task.description is not None:
					for description_line in task.description.splitlines():
						codes.append("# " + description_line)
				
				# Add actual code 
				args = list(map(convert_arg, task.args))
				kwargs = {k:convert_arg(arg) for k, arg in task.kwargs.items()}
				func_param = ', '.join(args + [k + "=" + arg for k, arg in kwargs.items()])

				for resource in task.return_values:
					rid_var_dict[resource.rid] = "v" + str(next_dummy_var_id)
					next_dummy_var_id += 1
					
				left_str = ", ".join([rid_var_dict[resource.rid] for resource in task.return_values])
				pipe = self.get_pipe(task.pid)
				pipename = obtain_pipename(pipe)
				codes.append(f"{left_str} = {pipename}({func_param})")
				
				graph_node_mapping[node] = textwrap.fill(f"{left_str} = {pipename}({func_param})", 60)
		if len(funcs) > 0:
			modules.add("dill")
		
		pre_codes = []
		pre_codes.append("# The codes are auto-generated from the RMS. Please refer to the original pipeline for details")
		pre_codes.append("###########")
		pre_codes.append("# Imports #")
		pre_codes.append("###########")
		for module in modules:
			pre_codes.append(f"import {module}")
		pre_codes.append("")
		pre_codes.append("########################")
		pre_codes.append("# Functions definition #")
		pre_codes.append("########################")
		for func_name, (hex_str, sourcecode) in funcs.items():
			pre_codes.append(f"{func_name} = dill.loads(bytes.fromhex(\"{hex_str}\"))")
			if sourcecode is not None:
				pre_codes.append("\n".join(map(lambda s: "# " + s, sourcecode.split("\n"))))
		pre_codes.append("")
		pre_codes.append("#############")
		pre_codes.append("# Pipelines #")
		pre_codes.append("#############")
		
		
		
		graph_node_mapping.update({(RMSEntryType.RESOURCE, k):v for k,v in rid_var_dict.items()})
		graph_node_mapping.update({(RMSEntryType.PIPE, k):v for k,v in redefined_pipes.items()})
		graph_node_mapping.update({node:self.get(node).file_path for node in g.nodes if node[0] == RMSEntryType.FILERESOURCE})
		return "\n".join(pre_codes + codes), g, graph_node_mapping
	
		
	def convert_to_rms_codes(self, rids=[], fids=[], tids=[]):
		import textwrap
		
		g = self.reconstruct_digraph(rids, fids, tids)
		graph_node_mapping = {}
		
		next_dummy_var_id = 1 
		next_dummy_func_id = 1
		rid_var_dict = {}
		funcs = dict()
		redefined_pipes = {}
		def obtain_pipename(pipe):
			nonlocal next_dummy_func_id
			if pipe.func.__module__ == "__main__":
				if pipe.pid not in redefined_pipes:
					if pipe.func.__name__ == "<lambda>":
						new_name = "f"
					else:
						new_name = pipe.func.__name__
					while new_name in funcs:
						new_name = new_name + str(next_dummy_func_id)
						next_dummy_func_id += 1
					funcs[new_name] = (dill.dumps(pipe.func).hex(), pipe.info["sourcecode"] if "sourcecode" in pipe.info else None)
					redefined_pipes[pipe.pid] = new_name
				pipename = redefined_pipes[pipe.pid]
			elif pipe.func.__module__ == "builtins":
				pipename = pipe.func.__name__
			else:
				modules.add(pipe.func.__module__)
				pipename = pipe.func.__module__ + "." + pipe.func.__name__
			return pipename
		
		def convert_arg(arg):
			if isinstance(arg, Resource):
				return rid_var_dict[arg.rid]
			elif isinstance(arg, FileResource):
				return "rms.file_from_path(" + json.dumps(arg.file_path) + ")"
			elif isinstance(arg, Pipe):
				#return f'rmsp.get_pipe("{arg.pid}")'
				return obtain_pipename(arg)
			else:
				return repr(arg)
				#return json.dumps(arg)
		codes = []
		modules = set()
		for node in nx.topological_sort(g):
			if node[0] == RMSEntryType.TASK:
				task = self.get_task(node[1])
				# Add description as comment
				if task.description is not None:
					for description_line in task.description.splitlines():
						codes.append("# " + description_line)
				
				# Add actual code 
				args = list(map(convert_arg, task.args))
				kwargs = {k:convert_arg(arg) for k, arg in task.kwargs.items()}
				func_param = ', '.join(args + [k + "=" + arg for k, arg in kwargs.items()])

				for resource in task.return_values:
					rid_var_dict[resource.rid] = "v" + str(next_dummy_var_id)
					next_dummy_var_id += 1
					
				left_str = ", ".join([rid_var_dict[resource.rid] for resource in task.return_values])
				pipe = self.get_pipe(task.pid)
				pipename = obtain_pipename(pipe)
				codes.append(f"{left_str} = {pipename}({func_param})")
				
				graph_node_mapping[node] = textwrap.fill(f"{left_str} = {pipename}({func_param})", 60)
		if len(funcs) > 0:
			modules.add("dill")
		
		pre_codes = []
		pre_codes.append("# The codes are auto-generated from the RMS. Please refer to the original pipeline for details")
		pre_codes.append("###########")
		pre_codes.append("# Imports #")
		pre_codes.append("###########")
		for module in modules:
			pre_codes.append(f"import {module}")
		pre_codes.append("")
		pre_codes.append("########################")
		pre_codes.append("# Functions definition #")
		pre_codes.append("########################")
		for func_name, (hex_str, sourcecode) in funcs.items():
			pre_codes.append(f"{func_name} = dill.loads(bytes.fromhex(\"{hex_str}\"))")
			if sourcecode is not None:
				pre_codes.append("\n".join(map(lambda s: "# " + s, sourcecode.split("\n"))))
		pre_codes.append("")
		pre_codes.append("#############")
		pre_codes.append("# Pipelines #")
		pre_codes.append("#############")
		
		
		
		graph_node_mapping.update({(RMSEntryType.RESOURCE, k):v for k,v in rid_var_dict.items()})
		graph_node_mapping.update({(RMSEntryType.PIPE, k):v for k,v in redefined_pipes.items()})
		graph_node_mapping.update({node:self.get(node).file_path for node in g.nodes if node[0] == RMSEntryType.FILERESOURCE})
		return "\n".join(pre_codes + codes), g, graph_node_mapping
			

		
	def wrap_as_function(self, fname, rids=[], fids=[], tids=[]):
		g = self.reconstruct_digraph(rids, fids, tids)
	
		next_dummy_var_id = 1 
		next_dummy_func_id = 1
		rid_var_dict = {}
		parameter_var_dict = {}
		file_dict = {}
		
		
		def convert_arg(arg):
			if isinstance(arg, Resource):
				return rid_var_dict[arg.rid]
			elif isinstance(arg, FileResource):
				return file_dict[arg.get_id()]
			elif isinstance(arg, Pipe):
				return f'rmsp.get_pipe("{arg.pid}")'
			else:
				pp = "z" + str(len(parameter_var_dict) + 1)
				parameter_var_dict[pp] = arg
				return pp
				#return repr(arg)
			
		for node, in_degree in g.in_degree:
			if in_degree == 0:
				if node[0] is RMSEntryType.FILERESOURCE:
					file_dict[node[1]] = "y" + str(len(file_dict) + 1)
					pass
				elif node[0] is RMSEntryType.RESOURCE:
					pass
				elif node[0] is RMSEntryType.TASK:
					pass
				else:
					raise Exception()
				
		codes = []
		modules = set()
		funcs = dict()
		redefined_pipes = {}
		for node in nx.topological_sort(g):
			if node[0] == RMSEntryType.TASK:
				task = self.get_task(node[1])
				# Add description as comment
				if task.description is not None:
					for description_line in task.description.splitlines():
						codes.append("# " + description_line)
	
				# Add actual code 
				args = list(map(convert_arg, task.args))
				kwargs = {k:convert_arg(arg) for k, arg in task.kwargs.items()}
				func_param = ', '.join(args + [k + "=" + arg for k, arg in kwargs.items()])
	
				for resource in task.return_values:
					rid_var_dict[resource.rid] = "v" + str(next_dummy_var_id)
					next_dummy_var_id += 1
				for fileresource in task.output_fileresources:
					file_dict[fileresource.fid] = "y" + str(len(file_dict) + 1)
				left_str = ", ".join([rid_var_dict[resource.rid] for resource in task.return_values])
				pipe = self.get_pipe(task.pid)
				if pipe.func.__module__ == "__main__":
					if pipe.pid not in redefined_pipes:
						if pipe.func.__name__ == "<lambda>":
							new_name = "f"
						else:
							new_name = pipe.func.__name__
						while new_name in funcs:
							new_name = new_name + str(next_dummy_func_id)
							next_dummy_func_id += 1
						funcs[new_name] = dill.dumps(pipe.func).hex()
						redefined_pipes[pipe.pid] = new_name
					pipename = redefined_pipes[pipe.pid]
				elif pipe.func.__module__ == "builtins":
					pipename = pipe.func.__name__
				else:
					modules.add(pipe.func.__module__)
					pipename = pipe.func.__module__ + "." + pipe.func.__name__
				codes.append(f"{left_str} = {pipename}({func_param})")
				
				
				if pipe.output_func is not None:
					left_str = ", ".join([file_dict[fileresource.fid] for fileresource in task.output_fileresources])
					modules.add(pipe.output_func.__module__)
					pipename = pipe.output_func.__module__ + "." + pipe.output_func.__name__
					codes.append(f"{left_str} = {pipename}({func_param})")
					
	
		if len(funcs) > 0:
			modules.add("dill")
	
		print("# The codes are auto-generated from the RMS. Please refer to the original pipeline for details")
		print("###########")
		print("# Imports #")
		print("###########")
		for module in modules:
			print(f"import {module}")
		print()
		print("########################")
		print("# Functions definition #")
		print("########################")
		for func_name, hex_str in funcs.items():
			print(f"{func_name} = dill.loads(bytes.fromhex(\"{hex_str}\"))")
		print()
		print("#############")
		print("# Pipelines #")
		print("#############")
		file_var_names = [f"{file_var_name}={repr(self.get_fileresource(fid).file_path)}" for fid, file_var_name in file_dict.items()]
		var_names = [f"{pp}={repr(arg)}" for pp, arg in parameter_var_dict.items()]
		func_input_parameter_str = ", ".join(file_var_names + var_names)
		print(f"def {fname}({func_input_parameter_str}):")
		for code in codes:
			print("\t" + code)
		print()
							

	
######################################################
# Experimental template
######################################################
class TemplateTask():
	def __init__(self, tid, func, output_func, return_volatile, arguments, return_values, output_files):
		self.tid = tid
		self.func = func
		self.output_func = output_func
		self.return_volatile = return_volatile
		self.arguments = arguments
		self.return_values = return_values
		self.output_files = output_files
class TemplateResource():
	def __init__(self, tid):
		self.tid = tid
class TemplateFunction():
	def __init__(self, func):
		self.func = func


def create_wrapped_func_codes(name, funcs, return_str):
	'''
	Create a function on the fly, by specifying the funcs. Function results are named as f{fidx}_r{ridx}. 
	:param funcs: A list of tuple (func, input_dict, output_value_count), where input_dict represents direct input into the function. 
	Keys of input_dict are argument names, and values are the results from other functions (e.g. f0_r0)
	
	'''
	def convert_param_default(d):
		if isinstance(d, str):
			return f"\"{d}\""
		elif isinstance(d, int):
			return f"{d}"
		elif isinstance(d, float):
			return f"{d}"
		else:
			return f"dill.loads(bytes.fromhex(\"{dill.dumps(parameter.default).hex()}\"))"
	def empty_default_string(parameter):
		if parameter.kind == inspect.Parameter.VAR_POSITIONAL:
			return "=[]"
		elif parameter.kind == inspect.Parameter.VAR_KEYWORD:
			return "={}"
		else:
			return ""
	# Parse parameters
	pipe_param_names = set()
	values_map = [dict() for _ in range(len(funcs))]
	pipe_parameters = []
	for func_idx, (func, input_dict, output_value_count) in enumerate(funcs):
		try:
			ba = inspect.signature(func)
		except ValueError:
			s = inspect.signature(func.__init__)
			ba = s.replace(parameters=[p for p in s.parameters.values() if p.name != "self"])
		parameters = ba.parameters
		for param_name in parameters:
			new_param_name = param_name
			param_dup_id = 0
			while new_param_name in pipe_param_names:
				param_dup_id += 1
				new_param_name = param_name + "_" + str(param_dup_id)
			pipe_param_names.add(new_param_name)
			if param_name in input_dict:
				values_map[func_idx][param_name] = input_dict[param_name] #f"f{input_dict[param_name][0]}_r{input_dict[param_name][1]}" 
			else:
				values_map[func_idx][param_name] = new_param_name
			if param_name in input_dict:
				continue
			if new_param_name == param_name:
				pipe_parameters.append(parameters[param_name])
			else:
				pipe_parameters.append(parameters[param_name].replace(name=new_param_name))
		
	modules = set()
	codes = []
	# Create the codes
	for func_idx, (func, input_dict, output_value_count) in enumerate(funcs):
		if func.__module__ == "builtins":
			pipename = func.__name__
		else:
			modules.add(func.__module__)
			pipename = func.__module__ + "." + func.__name__
		try:
			ba = inspect.signature(func)
		except ValueError:
			s = inspect.signature(func.__init__)
			ba = s.replace(parameters=[p for p in s.parameters.values() if p.name != "self"])
		parameters = ba.parameters
		args = []
		kwargs = {}
		for param_name, parameter in parameters.items():
			if parameter.kind == inspect.Parameter.POSITIONAL_ONLY:
				args.append(values_map[func_idx][param_name])
			elif parameter.kind == inspect.Parameter.VAR_POSITIONAL:
				args.append("*" + values_map[func_idx][param_name])
			elif parameter.kind == inspect.Parameter.VAR_KEYWORD:
				args.append("**" + values_map[func_idx][param_name])
			else:
				kwargs[param_name] = values_map[func_idx][param_name]
		func_param = ', '.join(args + [k + "=" + arg for k, arg in kwargs.items()])
		left_str = ", ".join(f"f{func_idx}_r{output_value_idx}" for output_value_idx in range(output_value_count))
		codes.append(f"{left_str} = {pipename}({func_param})")
	
	method_params_str = ",".join([parameter.name + (empty_default_string(parameter) if parameter.default == inspect._empty else ("=" + convert_param_default(parameter.default))) for parameter in sorted(pipe_parameters, key = lambda param: param.kind)])
	
	func_setup_codes = []
	func_setup_codes.append(f"def {name}({method_params_str}):")
	modules.add("dill")
	for module in modules:
		func_setup_codes.append(f"\timport {module}")
	for code in codes:
		func_setup_codes.append("\t" + code)
	func_setup_codes.append("\treturn " + return_str)
	final_codes = "\n".join(func_setup_codes)
	return final_codes
